"""Clustering from brain regions."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

import logging
from pathlib import Path

import dask.distributed
import networkx as nx
import numpy as np
import pandas as pd
import plotly.graph_objects as go
from bluepyparallel import evaluate
from bluepyparallel import init_parallel_factory
from plotly.subplots import make_subplots
from voxcell.math_utils import voxel_intersection

from axon_synthesis.constants import COORDS_COLS
from axon_synthesis.utils import build_layout_properties
from axon_synthesis.utils import disable_loggers
from axon_synthesis.utils import neurite_to_graph

logger = logging.getLogger(__name__)


def merge_similar_regions(regions, sub_segments):
    """Merge similar consecutive region IDs and corresponding sub-segments."""
    regions = np.array(regions)
    sub_segments = np.array(sub_segments)

    if regions.size == 0:
        return {
            "brain_regions": np.zeros_like([], dtype=int),
            "sub_segments": np.zeros_like([], shape=(0, 6), dtype=float),
        }

    # Find transitions between regions
    transitions = regions[:-1] != regions[1:]

    transition_indices = np.nonzero(transitions)[0]
    region_transition_indices = np.append(transition_indices, len(regions) - 1)
    seg_transition_indices = np.insert(region_transition_indices, 0, 0)

    # Build segments by region
    left_index = seg_transition_indices[:-1].copy()
    right_index = seg_transition_indices[1:].copy()
    couple_idx = np.vstack([left_index, right_index]).T
    segment_couples = sub_segments[couple_idx]

    region_sub_segments = np.hstack(
        [segment_couples[:, 0, [3, 4, 5]], segment_couples[:, 1, [3, 4, 5]]],
    )
    # Fix first start point
    region_sub_segments[0, [0, 1, 2]] = segment_couples[0, 0, [0, 1, 2]]

    # Remove segments with zero length
    seg_lengths = np.linalg.norm(
        region_sub_segments[:, [0, 1, 2]] - region_sub_segments[:, [3, 4, 5]],
        axis=1,
    )
    region_sub_segments = region_sub_segments[seg_lengths > 0]
    region_indices = regions[region_transition_indices][seg_lengths > 0]

    return {
        "brain_regions": region_indices,
        "sub_segments": region_sub_segments,
    }


def segment_region_ids(row, brain_regions):
    """Split segments according to brain regions and get the regions IDs."""
    start_pt = [row["source_x"], row["source_y"], row["source_z"]]
    end_pt = [row["target_x"], row["target_y"], row["target_z"]]
    indices, sub_segments = voxel_intersection(
        [start_pt, end_pt],
        brain_regions,
        return_sub_segments=True,
    )
    regions = brain_regions.raw[tuple(indices.T.tolist())]

    return merge_similar_regions(regions, sub_segments)


def cut_edges(edges, nodes, brain_regions, nb_workers, group_name):
    """Cut edges according to brain regions."""
    # Initialize Dask cluster
    with disable_loggers("asyncio", "distributed", "distributed.worker"):
        if nb_workers > 1:
            cluster = dask.distributed.LocalCluster(n_workers=nb_workers, timeout="60s")
            parallel_factory = init_parallel_factory("dask_dataframe", address=cluster)
        else:
            parallel_factory = init_parallel_factory(None)

        # Compute region indices of each segment
        all_brain_regions = evaluate(
            edges,
            segment_region_ids,
            [
                ["brain_regions", None],
                ["sub_segments", None],
            ],
            parallel_factory=parallel_factory,
            func_args=[brain_regions],
        )

        # Close the Dask cluster if opened
        if nb_workers > 1:
            parallel_factory.shutdown()
            cluster.close()

    logger.debug("%s: Computed brain regions for %s segments", group_name, len(edges))

    nb_intersected_regions = all_brain_regions["brain_regions"].apply(len)
    cut_edge_mask = nb_intersected_regions >= 2
    one_region_mask = nb_intersected_regions == 1

    # Set brain regions to not cut edges
    edges["brain_region"] = -1
    edges.loc[one_region_mask, "brain_region"] = all_brain_regions.loc[
        one_region_mask,
        "brain_regions",
    ].apply(lambda x: x[0])

    # Select edges that have to be cut
    edges_to_cut = edges.loc[cut_edge_mask].join(
        all_brain_regions.loc[cut_edge_mask, ["brain_regions", "sub_segments"]],
    )

    # Split lists into rows
    if not edges_to_cut.empty:
        region_sub_edges = edges_to_cut["brain_regions"].apply(pd.Series).stack().astype(int)
        segment_sub_edges = (
            edges_to_cut["sub_segments"].apply(lambda x: pd.Series(x.tolist())).stack()
        )

        region_sub_edges = region_sub_edges.reset_index(name="brain_region")
        segment_sub_edges = segment_sub_edges.reset_index(name="sub_segment")

        # Split coordinates into new columns
        segment_sub_edges[
            ["source_x", "source_y", "source_z", "target_x", "target_y", "target_z"]
        ] = pd.DataFrame(
            segment_sub_edges["sub_segment"].tolist(),
            columns=[
                "source_x",
                "source_y",
                "source_z",
                "target_x",
                "target_y",
                "target_z",
            ],
            index=segment_sub_edges.index,
        )

        # Join regions to sub-segments
        segment_sub_edges = segment_sub_edges.merge(region_sub_edges, on=["level_0", "level_1"])

        # Join sub-segments to edges to keep initial values for first and last segment points
        segment_sub_edges = segment_sub_edges.merge(
            edges[
                [
                    "source",
                    "target",
                    "source_is_terminal",
                    "target_is_terminal",
                    "source_section_id",
                    "target_section_id",
                    "source_sub_segment_num",
                    "target_sub_segment_num",
                    "source_is_intermediate_pt",
                    "target_is_intermediate_pt",
                ]
            ],
            left_on="level_0",
            right_index=True,
            how="left",
        )

        # Find indices of first and last element of each group
        # head_index = segment_sub_edges["level_1"] == 0
        # tail_index = segment_sub_edges.groupby("level_0").tail(1).index
        intermediate_sources = segment_sub_edges.groupby("level_0").tail(-1)
        intermediate_targets = segment_sub_edges.groupby("level_0").head(-1)

        # Add intermediate points to nodes
        intermediate_target_nodes = intermediate_targets[
            [
                "level_0",
                "level_1",
                "target_x",
                "target_y",
                "target_z",
                "target_section_id",
                "target_sub_segment_num",
            ]
        ].copy()

        intermediate_target_nodes["is_terminal"] = False  # By definition they can't be terminals
        intermediate_target_nodes["is_intermediate_pt"] = True  # Also by definition
        intermediate_target_nodes.rename(
            columns={
                "target_x": "x",
                "target_y": "y",
                "target_z": "z",
                "target_section_id": "section_id",
                "target_sub_segment_num": "sub_segment_num",
            },
            inplace=True,
        )

        intermediate_target_nodes.reset_index(inplace=True)
        intermediate_target_nodes.index += nodes.index.max() + 1
        intermediate_target_nodes["new_index"] = intermediate_target_nodes.index

        # Set source and target indices
        segment_sub_edges.loc[intermediate_targets.index, "target"] = (
            segment_sub_edges.loc[intermediate_targets.index]
            .merge(
                intermediate_target_nodes[["level_0", "level_1", "new_index"]],
                on=["level_0", "level_1"],
            )["new_index"]
            .tolist()
        )
        segment_sub_edges.loc[intermediate_sources.index, "source"] = segment_sub_edges.loc[
            intermediate_targets.index,
            "target",
        ].tolist()

        # Set source_is_terminal and target_is_terminal attributes to False (the intermediate
        # points can not be terminals)
        segment_sub_edges.loc[intermediate_sources.index, "source_is_terminal"] = False
        segment_sub_edges.loc[intermediate_targets.index, "target_is_terminal"] = False

        # Fix sub-segment numbers
        # TODO: Vectorize the for loop
        intermediate_target_nodes["idx_shift"] = intermediate_target_nodes.groupby(
            "section_id",
        ).cumcount()
        intermediate_target_nodes["sub_segment_num"] += intermediate_target_nodes["idx_shift"]
        for i in intermediate_target_nodes.itertuples():
            nodes.loc[
                (nodes["section_id"] == i.section_id)
                & (nodes["sub_segment_num"] >= i.sub_segment_num),
                "sub_segment_num",
            ] += 1
        segment_sub_edges.loc[
            intermediate_sources.index,
            ["source_sub_segment_num", "target_sub_segment_num"],
        ] += 1

        # Build new DataFrames
        new_nodes = pd.concat(
            [
                nodes,
                intermediate_target_nodes[
                    [i for i in nodes.columns if i in intermediate_target_nodes.columns]
                ],
            ],
        )

        new_edges = pd.concat(
            [
                edges.drop(np.where(cut_edge_mask.values)[0]),
                segment_sub_edges[edges.columns],
            ],
            ignore_index=True,
        )
    else:
        new_nodes = nodes
        new_edges = edges

    return new_nodes, new_edges


def _find_wm_first_nested_region(region_id, wm_regions, region_map) -> int:
    """Find the first nested region ID that is available in the white matter recipe."""
    if region_id in wm_regions or region_id <= 0:
        return region_id

    ids = region_map.get(region_id, attr="id", with_ascendants=True)
    for i in ids[1:]:
        if i in wm_regions:
            return i

    return region_id


def compute_clusters(  # noqa: PLR0913
    atlas,
    wmr,
    config,
    config_name,
    axon,
    axon_id,
    group_name,
    group_path,
    group,
    nb_workers,
    tuft_morphologies_path,
    figure_path,
    *,
    debug=False,
    **_kwargs,
):
    """Gather points according to connected components in brain regions."""
    nodes, edges, _ = neurite_to_graph(axon, keep_section_segments=True)

    nodes["is_intermediate_pt"] = (
        nodes["sub_segment_num"]
        != nodes.merge(
            nodes.groupby("section_id")["sub_segment_num"].max(),
            left_on="section_id",
            right_index=True,
            suffixes=("", "_max"),
        )["sub_segment_num_max"]
    )

    edges = edges.join(nodes.add_prefix("source_"), on="source")
    edges = edges.join(nodes.add_prefix("target_"), on="target")

    # Cut edges according to brain regions
    new_nodes, new_edges = cut_edges(edges, nodes, atlas.brain_regions, nb_workers, group_name)

    # Normalize node indices
    new_nodes.sort_values(["section_id", "sub_segment_num"], inplace=True)
    new_nodes.index.name = "old_index"
    new_nodes.reset_index(inplace=True)
    new_nodes.index -= 1
    new_nodes.index.name = "id"
    new_nodes["new_index"] = new_nodes.index

    new_edges["source"] = new_edges.merge(
        new_nodes[["old_index", "new_index"]],
        left_on="source",
        right_on="old_index",
        how="left",
    )["new_index"]
    new_edges["target"] = new_edges.merge(
        new_nodes[["old_index", "new_index"]],
        left_on="target",
        right_on="old_index",
        how="left",
    )["new_index"]
    new_nodes.drop(columns=["new_index", "old_index"], inplace=True)

    # Normalize nested brain regions
    wm_regions = np.sort(wmr.populations["atlas_region_id"].unique())

    new_edges["wm_brain_region"] = new_edges["brain_region"].apply(
        _find_wm_first_nested_region,
        args=(wm_regions, atlas.region_map),
    )

    brain_region_attr = "wm_brain_region" if config["wm_unnesting"] else "brain_region"

    # Create a graph from these new nodes and edges
    graph = nx.from_pandas_edgelist(new_edges, create_using=nx.Graph)
    nx.set_node_attributes(graph, new_nodes.to_dict("index"))
    nx.set_edge_attributes(
        graph,
        new_edges.set_index(["source", "target"]).to_dict("index"),
    )

    # Get subgraphs of each brain region
    region_sub_graphs = {
        brain_region: graph.edge_subgraph(
            edges_component[["source", "target"]].to_records(index=False).tolist(),
        )
        for brain_region, edges_component in new_edges.groupby(brain_region_attr)
    }

    # Get connected components of each brain region
    region_components = {
        brain_region: list(nx.connected_components(region_sub_graphs[brain_region]))
        for brain_region, sub_graph in region_sub_graphs.items()
    }

    region_component_subgraphs = {
        brain_region: [region_sub_graphs[brain_region].subgraph(comp) for comp in components]
        for brain_region, components in region_components.items()
    }

    region_acronyms = {
        brain_region: atlas.region_map.get(brain_region, attr="acronym")
        for brain_region in region_components
        if brain_region > 0
    }

    # Create a cluster ID for each component
    group_nodes = (
        group.reset_index()
        .merge(
            new_nodes.loc[~new_nodes["is_intermediate_pt"]].reset_index()[["section_id", "id"]],
            on="section_id",
        )
        .set_index("index")
    )
    group_nodes = (
        group_nodes.reset_index()
        .merge(
            new_edges[["target", brain_region_attr]],
            left_on="id",
            right_on="target",
        )
        .set_index("index")
    )

    cluster_id = 0
    # cluster_ids = {}
    for components in region_components.values():  # pylint: disable=unused-variable
        for component in components:
            group_nodes.loc[group_nodes["id"].isin(list(component)), "tuft_id"] = cluster_id
            cluster_id = group_nodes["tuft_id"].max() + 1

    group["tuft_id"] = group_nodes["tuft_id"]

    new_terminal_points = []
    # new_terminal_id = group["terminal_id"].max() + 1
    for new_terminal_id, (cluster_id, i) in enumerate(group.groupby("tuft_id")):
        new_terminal_points.append(
            [
                group_name,
                group_path,
                config_name,
                axon_id,
                new_terminal_id if cluster_id != -1 else 0,
                len(i),
                *i[COORDS_COLS].mean().tolist(),
            ],
        )

    if debug:
        plot(
            region_component_subgraphs,
            region_acronyms,
            str(figure_path / f"{group_name}_region_clusters.html"),
        )

        tuft_brain_region_path = tuft_morphologies_path / f"{group_name}_{axon_id}.csv"
        logger.debug("Export tuft brain regions to %s", tuft_brain_region_path)
        group_nodes["region_acronym"] = group_nodes["wm_brain_region"].map(region_acronyms)
        # group_nodes["tuft_morph_path"] = group_nodes.apply(
        #     lambda row: task.tuft_morph_path(group_name, axon_id, row["tuft_id"]), axis=1
        # )
        # group_nodes.loc[group_nodes["tuft_id"] == 0, "tuft_morph_path"] = None
        group_nodes.to_csv(tuft_brain_region_path)

    return new_terminal_points, group["tuft_id"]


def plot(region_component_subgraphs, region_acronyms, filepath):
    """Plot the resulting nodes and edges."""
    rng = np.random.default_rng()

    total_num = sum(len(i) for i in region_component_subgraphs.values())
    all_colors = np.arange(total_num).tolist()
    rng.shuffle(all_colors)
    x = []
    y = []
    z = []
    color = []
    acronym = []
    # annotations = []
    for region, subgraphs in region_component_subgraphs.items():
        acr = region_acronyms.get(region, "UNKNOWN")
        for subgraph in subgraphs:
            tmp = all_colors.pop()
            for start_node, end_node in subgraph.edges:
                edge_data = subgraph.get_edge_data(start_node, end_node)
                x.extend(
                    [
                        edge_data["source_x"],
                        edge_data["target_x"],
                        None,
                    ],
                )
                y.extend(
                    [
                        edge_data["source_y"],
                        edge_data["target_y"],
                        None,
                    ],
                )
                z.extend(
                    [
                        edge_data["source_z"],
                        edge_data["target_z"],
                        None,
                    ],
                )
                color.extend([tmp, tmp, tmp])
                acronym.extend([acr, acr, acr])
    edge_trace = go.Scatter3d(
        x=x,
        y=y,
        z=z,
        mode="lines",
        line={
            "color": color,
            "width": 4,
            "colorscale": "hsv",
        },
        name="Morphology nodes to cluster",
        hovertext=acronym,
    )

    fig = make_subplots(
        cols=1,
        specs=[[{"is_3d": True}]],
        subplot_titles=("Region clusters"),
    )

    fig.add_trace(edge_trace, row=1, col=1)

    layout_props = build_layout_properties(np.stack([x, y, z]).T, 0.1)

    fig.update_scenes(layout_props)
    fig.update_layout(title=Path(filepath).stem)

    # Export figure
    Path(filepath).parent.mkdir(parents=True, exist_ok=True)
    fig.write_html(filepath)



"""Some plot utils for clustering."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

from pathlib import Path

import numpy as np
import plotly.graph_objects as go
from matplotlib import pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from morph_tool import resampling
from neurom.core import Morphology
from plotly.subplots import make_subplots
from plotly_helper.neuron_viewer import NeuronBuilder

from axon_synthesis.constants import COMMON_ANCESTOR_COORDS_COLS
from axon_synthesis.constants import COORDS_COLS
from axon_synthesis.utils import add_camera_sync
from axon_synthesis.utils import build_layout_properties
from axon_synthesis.utils import disable_loggers


def plot_clusters(morph, clustered_morph, group, group_name, cluster_df, output_path):
    """Plot clusters to a HTML figure."""
    plotted_morph = Morphology(
        resampling.resample_linear_density(morph, 0.005),
        name=Path(group_name).with_suffix("").name,
    )
    fig_builder = NeuronBuilder(plotted_morph, "3d", line_width=4, title=f"{plotted_morph.name}")

    x, y, z = group[COORDS_COLS].to_numpy().T
    node_trace = go.Scatter3d(
        x=x,
        y=y,
        z=z,
        mode="markers",
        marker={"size": 3, "color": "black"},
        name="Morphology nodes",
    )
    x, y, z = cluster_df[COORDS_COLS].to_numpy().T
    cluster_trace = go.Scatter3d(
        x=x,
        y=y,
        z=z,
        mode="markers",
        marker={"size": 5, "color": "chocolate"},
        name="Cluster centers",
    )
    cluster_lines = [
        [
            [
                i["x"],
                cluster_df.loc[cluster_df["terminal_id"] == i["tuft_id"], "x"].iloc[0],
                None,
            ],
            [
                i["y"],
                cluster_df.loc[cluster_df["terminal_id"] == i["tuft_id"], "y"].iloc[0],
                None,
            ],
            [
                i["z"],
                cluster_df.loc[cluster_df["terminal_id"] == i["tuft_id"], "z"].iloc[0],
                None,
            ],
        ]
        for i in group.to_dict("records")
        if i["tuft_id"] >= 0
    ]
    edge_trace = go.Scatter3d(
        x=[j for i in cluster_lines for j in i[0]],
        y=[j for i in cluster_lines for j in i[1]],
        z=[j for i in cluster_lines for j in i[2]],
        hoverinfo="none",
        mode="lines",
        line={
            "color": "green",
            "width": 4,
        },
        name="Morphology nodes to cluster",
    )

    # Build the clustered morph figure
    clustered_builder = NeuronBuilder(
        clustered_morph,
        "3d",
        line_width=4,
        title=f"Clustered {clustered_morph.name}",
    )

    # Create the figure from the traces
    fig = make_subplots(
        cols=2,
        specs=[[{"is_3d": True}, {"is_3d": True}]],
        subplot_titles=("Node clusters", "Clustered morphology"),
    )

    morph_data = fig_builder.get_figure()["data"]
    fig.add_traces(morph_data, rows=[1] * len(morph_data), cols=[1] * len(morph_data))
    fig.add_trace(node_trace, row=1, col=1)
    fig.add_trace(edge_trace, row=1, col=1)
    fig.add_trace(cluster_trace, row=1, col=1)

    clustered_morph_data = clustered_builder.get_figure()["data"]
    fig.add_traces(
        clustered_morph_data,
        rows=[1] * len(clustered_morph_data),
        cols=[2] * len(clustered_morph_data),
    )
    cluster_trace.showlegend = False
    fig.add_trace(cluster_trace, row=1, col=2)

    layout_props = build_layout_properties(morph.points, 0.1)

    fig.update_scenes(layout_props)
    fig.update_layout(title=group_name)

    # Export figure
    Path(output_path).parent.mkdir(parents=True, exist_ok=True)
    fig.write_html(str(output_path))

    add_camera_sync(str(output_path))


@disable_loggers(
    "matplotlib.font_manager",
    "matplotlib.pyplot",
    "matplotlib.backends.backend_pdf",
    "matplotlib.ticker",
    "PIL.PngImagePlugin",
)
def plot_cluster_properties(cluster_props_df, output_path):
    """Plot the cluster properties to a PDF figure."""
    if cluster_props_df.empty:
        return

    with PdfPages(str(output_path)) as pdf:
        ax = cluster_props_df.plot.scatter(
            x="path_distance",
            y="size",
            title="Cluster size vs path distance",
            legend=True,
        )
        ax.set_yscale("log")
        pdf.savefig()
        plt.close()

        ax = cluster_props_df.plot.scatter(
            x="radial_distance",
            y="size",
            title="Cluster size vs radial distance",
            legend=True,
        )
        ax.set_yscale("log")
        pdf.savefig()
        plt.close()

        ax = (
            plt.scatter(
                x=cluster_props_df["radial_distance"],
                y=np.linalg.norm(
                    np.stack(cluster_props_df["center_coords"].to_numpy())
                    - cluster_props_df[COMMON_ANCESTOR_COORDS_COLS].to_numpy(),
                    axis=1,
                ),
            )
            .get_figure()
            .gca()  # type: ignore[union-attr]
        )
        ax.set_title("Cluster radial length vs radial distance")
        pdf.savefig()
        plt.close()

        ax = cluster_props_df.plot.scatter(
            x="size",
            y="path_length",
            title="Path length vs cluster size",
            legend=True,
        )
        pdf.savefig()
        plt.close()

        ax = cluster_props_df.plot.scatter(
            x="path_distance",
            y="path_length",
            title="Path length vs path distance",
            legend=True,
        )
        pdf.savefig()
        plt.close()

        ax = cluster_props_df.plot.scatter(
            x="radial_distance",
            y="path_length",
            title="Path length vs radial distance",
            legend=True,
        )
        pdf.savefig()
        plt.close()



"""Tests for the clustering processing functions."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

import numpy as np
import pytest

from axon_synthesis.inputs import clustering


class TestFromBrainRegion:
    """Test functions from `axon_synthesis.main_trunk.clustering.from_brain_regions`."""

    @pytest.mark.parametrize(
        ("regions", "sub_segments", "expected_regions", "expected_sub_segments"),
        [
            pytest.param(
                [1, 2],
                [[0, 0, 0, 0.5, 0, 0], [0.5, 0, 0, 1, 0, 0]],
                [1, 2],
                [[0, 0, 0, 0.5, 0, 0], [0.5, 0, 0, 1, 0, 0]],
                id="two_single_regions",
            ),
            pytest.param(
                [1, 1, 2],
                [[0, 0, 0, 0.25, 0, 0], [0.25, 0, 0, 0.5, 0, 0], [0.5, 0, 0, 1, 0, 0]],
                [1, 2],
                [[0, 0, 0, 0.5, 0, 0], [0.5, 0, 0, 1, 0, 0]],
                id="two_regions_with_first_duplicated",
            ),
            pytest.param(
                [1, 2, 2],
                [[0, 0, 0, 0.5, 0, 0], [0.5, 0, 0, 0.75, 0, 0], [0.75, 0, 0, 1, 0, 0]],
                [1, 2],
                [[0, 0, 0, 0.5, 0, 0], [0.5, 0, 0, 1, 0, 0]],
                id="two_regions_with_second_duplicated",
            ),
            pytest.param(
                [1, 1, 2, 2],
                [
                    [0, 0, 0, 0.25, 0, 0],
                    [0.25, 0, 0, 0.5, 0, 0],
                    [0.5, 0, 0, 0.75, 0, 0],
                    [0.75, 0, 0, 1, 0, 0],
                ],
                [1, 2],
                [[0, 0, 0, 0.5, 0, 0], [0.5, 0, 0, 1, 0, 0]],
                id="two_regions_with_both_duplicated",
            ),
            pytest.param(
                [1, 2, 3],
                [
                    [0, 0, 0, 0.5, 0, 0],
                    [0.5, 0, 0, 0.75, 0, 0],
                    [0.75, 0, 0, 1, 0, 0],
                ],
                [1, 2, 3],
                [[0, 0, 0, 0.5, 0, 0], [0.5, 0, 0, 0.75, 0, 0], [0.75, 0, 0, 1, 0, 0]],
                id="three_single_regions",
            ),
            pytest.param(
                [1, 1, 2, 3],
                [
                    [0, 0, 0, 0.25, 0, 0],
                    [0.25, 0, 0, 0.5, 0, 0],
                    [0.5, 0, 0, 0.75, 0, 0],
                    [0.75, 0, 0, 1, 0, 0],
                ],
                [1, 2, 3],
                [[0, 0, 0, 0.5, 0, 0], [0.5, 0, 0, 0.75, 0, 0], [0.75, 0, 0, 1, 0, 0]],
                id="three_regions_first_duplicated",
            ),
            pytest.param(
                [1, 1, 2, 3],
                [
                    [0, 0, 0, 0.25, 0, 0],
                    [0.25, 0, 0, 0.5, 0, 0],
                    [0.5, 0, 0, 0.5, 0, 0],
                    [0.5, 0, 0, 1, 0, 0],
                ],
                [1, 3],
                [[0, 0, 0, 0.5, 0, 0], [0.5, 0, 0, 1, 0, 0]],
                id="three_regions_first_duplicated_second_0_length",
            ),
            pytest.param(
                [1, 1],
                [[0, 0, 0, 0.5, 0, 0], [0.5, 0, 0, 1, 0, 0]],
                [1],
                [[0, 0, 0, 1, 0, 0]],
                id="one_duplicated_region",
            ),
            pytest.param(
                [1, 1],
                [[0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0]],
                [],
                [],
                id="one_duplicated_region_0_length",
            ),
            pytest.param(
                [],
                [],
                [],
                [],
                id="empty_region_list",
            ),
            pytest.param(
                [1],
                [[0, 0, 0, 1, 0, 0]],
                [1],
                [[0, 0, 0, 1, 0, 0]],
                id="one_single_region",
            ),
        ],
    )
    @pytest.mark.parametrize(
        "reverse",
        [pytest.param(True, id="reversed"), pytest.param(False, id="regular")],
    )
    def test_merge_similar_regions(
        self,
        regions,
        sub_segments,
        expected_regions,
        expected_sub_segments,
        reverse,
    ):
        """Test that consecutive regions are properly merged."""
        regions = np.array(regions)
        sub_segments = np.array(sub_segments)

        # Prepare data for reversed case
        if reverse and len(regions) > 0:
            regions = regions[::-1]
            sub_segments = sub_segments[::-1, [3, 4, 5, 0, 1, 2]]
            expected_regions = expected_regions[::-1]
            if len(expected_sub_segments) > 0:
                expected_sub_segments = (
                    np.array(expected_sub_segments)[::-1, [3, 4, 5, 0, 1, 2]]
                ).tolist()

        # Call merge_similar_regions()
        res = clustering.from_brain_regions.merge_similar_regions(
            regions.tolist(),
            sub_segments.tolist(),
        )

        # Handle empty case
        if len(expected_sub_segments) == 0:
            expected_sub_segments = np.zeros_like([], shape=(0, 6), dtype=float)

        # Check results
        np.testing.assert_array_equal(res["brain_regions"], expected_regions)
        np.testing.assert_array_equal(res["sub_segments"], expected_sub_segments)

    class TestSegmentRegionIds:
        """Test that the segments are properly cut by brain regions."""

        def test_zero(self, brain_regions):
            """Test with 0-length segment."""
            seg = {
                "source": 0,
                "source_x": 0,
                "source_y": 0,
                "source_z": 0,
                "target": 1,
                "target_x": 0,
                "target_y": 0,
                "target_z": 0,
            }
            res = clustering.from_brain_regions.segment_region_ids(seg, brain_regions)
            assert res["brain_regions"].tolist() == []
            assert res["sub_segments"].tolist() == []

        def test_seg_in_one_region(self, brain_regions):
            """Test with a segment that is completely contained in a region."""
            seg = {
                "source": 0,
                "source_x": 0,
                "source_y": 0,
                "source_z": 0,
                "target": 1,
                "target_x": 10,
                "target_y": 20,
                "target_z": 0,
            }
            res = clustering.from_brain_regions.segment_region_ids(seg, brain_regions)
            np.testing.assert_array_equal(res["brain_regions"], [20])
            np.testing.assert_array_equal(res["sub_segments"], [[0, 0, 0, 10, 20, 0]])

        def test_seg_in_two_regions_one_voxel_both(self, brain_regions):
            """Test with a segment crossing 2 regions, but only one voxel in each."""
            seg = {
                "source": 0,
                "source_x": -10,
                "source_y": 0,
                "source_z": 0,
                "target": 1,
                "target_x": 10,
                "target_y": 0,
                "target_z": 0,
            }
            res = clustering.from_brain_regions.segment_region_ids(seg, brain_regions)
            np.testing.assert_array_equal(res["brain_regions"], [2, 20])
            np.testing.assert_array_equal(
                res["sub_segments"],
                [[-10, 0, 0, 0, 0, 0], [0, 0, 0, 10, 0, 0]],
            )

        def test_seg_in_two_regions_one_voxel_start(self, brain_regions):
            """Test with a segment crossing 2 regions, but only one voxel in start region."""
            seg = {
                "source": 0,
                "source_x": -10,
                "source_y": 0,
                "source_z": 0,
                "target": 1,
                "target_x": 500,
                "target_y": 0,
                "target_z": 0,
            }
            res = clustering.from_brain_regions.segment_region_ids(seg, brain_regions)
            np.testing.assert_array_equal(res["brain_regions"], [2, 20])
            np.testing.assert_array_equal(
                res["sub_segments"],
                [[-10, 0, 0, 0, 0, 0], [0, 0, 0, 500, 0, 0]],
            )

        def test_seg_in_two_regions_one_voxel_end(self, brain_regions):
            """Test with a segment crossing 2 regions, but only one voxel in end region."""
            seg = {
                "source": 0,
                "source_x": -500,
                "source_y": 0,
                "source_z": 0,
                "target": 1,
                "target_x": 10,
                "target_y": 0,
                "target_z": 0,
            }
            res = clustering.from_brain_regions.segment_region_ids(seg, brain_regions)
            np.testing.assert_array_equal(res["brain_regions"], [2, 20])
            np.testing.assert_array_equal(
                res["sub_segments"],
                [[-500, 0, 0, 0, 0, 0], [0, 0, 0, 10, 0, 0]],
            )

        def test_seg_in_two_regions_several_voxels(self, brain_regions):
            """Test with a segment crossing 2 regions on several voxel in each."""
            seg = {
                "source": 0,
                "source_x": -500,
                "source_y": 0,
                "source_z": 0,
                "target": 1,
                "target_x": 500,
                "target_y": 0,
                "target_z": 0,
            }
            res = clustering.from_brain_regions.segment_region_ids(seg, brain_regions)
            np.testing.assert_array_equal(res["brain_regions"], [2, 20])
            np.testing.assert_array_equal(
                res["sub_segments"],
                [[-500, 0, 0, 0, 0, 0], [0, 0, 0, 500, 0, 0]],
            )



"""Entries of the Command Line Interface dedicated to the input creation."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
# LICENSE HEADER MANAGED BY add-license-header

from pathlib import Path

import click
from click_option_group import optgroup

from axon_synthesis import inputs
from axon_synthesis.cli.common import atlas_kwargs_to_config
from axon_synthesis.cli.common import atlas_options
from axon_synthesis.cli.common import parallel_kwargs_to_config
from axon_synthesis.cli.common import parallel_options
from axon_synthesis.cli.utils import DictParam
from axon_synthesis.cli.utils import GlobalConfig
from axon_synthesis.white_matter_recipe import WmrConfig
from axon_synthesis.white_matter_recipe import fetch


def wmr_kwargs_to_config(config) -> None:
    """Extract the atlas arguments from given config to create an AtlasConfig object."""
    wmr_path = config.pop("wmr_path", None)
    wmr_subregion_uppercase = config.pop("wmr_subregion_uppercase")
    wmr_subregion_remove_prefix = config.pop("wmr_subregion_remove_prefix")
    wmr_sub_region_separator = config.pop("wmr_sub_region_separator")
    if wmr_path is None:
        config["wmr_config"] = None
    else:
        config["wmr_config"] = WmrConfig(
            wmr_path,
            wmr_subregion_uppercase,
            wmr_subregion_remove_prefix,
            wmr_sub_region_separator,
        )


clustering_parameters_option = click.option(
    "--clustering-parameters",
    type=DictParam(),
    required=True,
    help="Parameters used for the clustering algorithm",
)


@click.command(short_help="Fetch the White Matter Recipe file from a given repository")
@click.option(
    "--url",
    type=str,
    required=True,
    help="The URL of the repository that contains the target file",
)
@click.option(
    "--file-path",
    type=str,
    default="white_matter_FULL_RECIPE_v1p20.yaml",
    help="The path of the target file in the target repository",
)
@click.option(
    "--version-reference",
    type=str,
    help=(
        "The version that should be used in the repository (can be a tag, a commit hash or a "
        "branch name)"
    ),
)
@click.option(
    "-o",
    "--output-path",
    type=click.Path(path_type=Path),
    required=True,
    help="The path to the destination file",
)
def fetch_white_matter_recipe(**kwargs):
    """Command to fetch the White Matter Recipe from a git repository."""
    fetch(**kwargs)


@click.command(
    short_help=(
        "Generate all the parameters from the Atlas, the White Matter Recipe and the input "
        "morphologies"
    ),
)
@click.option(
    "--morphology-dir",
    type=click.Path(exists=True, file_okay=False),
    required=True,
    help="The directory containing the input morphologies",
)
@atlas_options(required=False)
@optgroup.group(
    "White Matter Recipe parameters",
    help="Parameters used to load and process the White Matter Recipe file",
)
@optgroup.option(
    "--wmr-path",
    type=click.Path(exists=True, dir_okay=False),
    # required=False,
    help="Path to the White Matter Recipe file",
)
@optgroup.option(
    "--wmr-subregion-uppercase",
    is_flag=True,
    default=False,
    help="",
)
@optgroup.option(
    "--wmr-subregion-keep-prefix",
    "wmr_subregion_remove_prefix",
    flag_value=True,
    default=True,
    help="",
)
@optgroup.option(
    "--wmr-subregion-remove-prefix",
    "wmr_subregion_remove_prefix",
    flag_value=False,
    help="",
)
@optgroup.option(
    "--wmr-sub-region-separator",
    type=str,
    default="",
    help="",
)
@clustering_parameters_option
@click.option(
    "--neuron-density",
    type=click.FloatRange(min=0, min_open=True),
    default=1e-2,
    help=(
        "The density of neurons in the atlas (we suppose here that this density is uniform). "
        "This density should be given in number of neurons by cube atlas-unit (usually "
        "micrometer)."
    ),
)
@click.option(
    "--bouton-density",
    type=click.FloatRange(min=0, min_open=True),
    default=0.2,
    help="The density of boutons along the axons (we suppose here that this density is uniform).",
)
@click.option(
    "-o",
    "--output-dir",
    type=click.Path(path_type=Path),
    required=True,
    help="Output directory",
)
@parallel_options
@click.pass_obj
def create_inputs(global_config: GlobalConfig, **kwargs):
    """The command to create inputs."""
    global_config.to_config(kwargs)
    atlas_kwargs_to_config(kwargs)
    wmr_kwargs_to_config(kwargs)
    parallel_kwargs_to_config(kwargs)
    inputs.create.create_inputs(**kwargs)



"""Tests for the post-processing functions."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

import matplotlib.pyplot as plt
import numpy as np
import numpy.testing as npt

from axon_synthesis.synthesis.main_trunk import post_process

from . import use_matplotlib_backend


class TestRandomWalk:
    """Test random walk functions."""

    def test_simple(self, interactive_plots):
        """Test that the random walk works properly."""
        start_pt = np.array([0, 0, 0])
        intermediate_pts = np.array(
            [
                [1, 0, 0],
                [1, 1, 0],
                [0, 1, 0],
                [0, 2, 0],
            ],
        )
        length_stats = {
            "norm": 0.05,
            "std": 0.01,
        }
        # angle_stats = {
        #     "norm": 30,
        #     "std": 10,
        # }
        rng = np.random.default_rng(0)

        config = post_process.PostProcessConfig(
            history_path_length=None,
            global_target_coeff=0,
            target_coeff=2,
            random_coeff=2,
            history_coeff=2,
        )

        points, (latest_lengths, latest_directions) = post_process.random_walk(
            start_pt,
            intermediate_pts,
            length_stats,
            # angle_stats,
            config=config,
            previous_history=([length_stats["norm"]] * 10, [[0, -1, 0]] * 10),
            rng=rng,
        )

        # ################################################################## #
        # Plot before testing the results
        if interactive_plots:
            with use_matplotlib_backend("QtAgg"):
                fig = plt.figure()
                ax = fig.add_subplot(projection="3d")
                ax.plot(
                    points[:, 0],
                    points[:, 1],
                    points[:, 2],
                    label="Random walk",
                )
                ax.scatter(*start_pt, label="Start point")
                ax.scatter(
                    intermediate_pts[:, 0],
                    intermediate_pts[:, 1],
                    intermediate_pts[:, 2],
                    label="Intermediate targets",
                )
                for num, i in enumerate(intermediate_pts):
                    ax.text(i[0], i[1], i[2], str(num), size=20, zorder=1, color="k")  # type: ignore[arg-type]
                ax.legend()
                ax.set(xlim3d=(-0.5, 1.5), xlabel="X")
                ax.set(ylim3d=(-0.5, 1.5), ylabel="Y")
                ax.set(zlim3d=(-0.5, 1.5), zlabel="Z")
                plt.show()
        # ################################################################## #

        assert len(points) == 93
        npt.assert_array_equal(points[0], [0, 0, 0])
        npt.assert_array_equal(points[-1], [0, 2, 0])
        npt.assert_array_almost_equal(points[-2], [-1.3573e-2, 2.005857, 0.018934])
        assert len(latest_lengths) == 5
        assert len(latest_directions) == 5



"""Find the target points of the input morphologies."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

import itertools
import logging
from typing import Any

import numpy as np
import pandas as pd
from h5py import File
from numpy.random import Generator
from scipy.spatial import KDTree

from axon_synthesis.atlas import AtlasHelper
from axon_synthesis.constants import COORDS_COLS
from axon_synthesis.constants import DEFAULT_POPULATION
from axon_synthesis.constants import SOURCE_COORDS_COLS
from axon_synthesis.constants import TARGET_COORDS_COLS
from axon_synthesis.typing import FileType
from axon_synthesis.typing import SeedType
from axon_synthesis.utils import ignore_warnings

LOGGER = logging.getLogger(__name__)


def compute_coords(
    target_points: pd.DataFrame,
    brain_regions_masks: File | None,
    rng: Generator,
    *,
    atlas: AtlasHelper | None = None,
) -> None:
    """Compute the target coordinates if they are missing."""
    if set(TARGET_COORDS_COLS).difference(target_points.columns):
        if brain_regions_masks is not None:
            mask_tmp = (
                target_points.loc[~target_points["target_brain_region_id"].isna()]
                .sort_values("target_brain_region_id")
                .index
            )
            target_points.loc[:, TARGET_COORDS_COLS] = np.nan

            def get_coords(group: pd.DataFrame) -> np.ndarray:
                try:
                    return rng.choice(  # type: ignore[arg-type, return-value]
                        brain_regions_masks[str(int(group.name))][:], size=len(group)
                    )
                except Exception as e:
                    logging.warning("Error %s for target_brain_region_id %s", repr(e), group.name)
                    return np.repeat(np.nan, len(group))

            target_points.loc[mask_tmp, TARGET_COORDS_COLS] = (
                target_points.groupby("target_brain_region_id")
                .apply(get_coords)  # type: ignore[arg-type]
                .explode()
                .sort_index()
                .apply(pd.Series)
                .to_numpy()
            )
        else:
            msg = (
                f"The target points should contain the {TARGET_COORDS_COLS} columns when no brain "
                "region mask is given"
            )
            raise RuntimeError(msg)
        if atlas is not None:
            # Convert indices into coordinates
            target_points.loc[:, TARGET_COORDS_COLS] = atlas.brain_regions.indices_to_positions(
                target_points[TARGET_COORDS_COLS].to_numpy()  # noqa: RUF005
                + [0.5, 0.5, 0.5]
            ) + atlas.get_random_voxel_shifts(len(target_points), rng=rng)
            # place the targets in the correct hemisphere, if hemisphere is given
            if "target_hemisphere" in target_points.columns:
                target_points[TARGET_COORDS_COLS] = target_points.apply(  # type: ignore[call-overload]
                    lambda row: atlas.place_point_in_hemisphere(
                        row[TARGET_COORDS_COLS], row["target_hemisphere"]
                    ),
                    axis=1,
                )[TARGET_COORDS_COLS]


def select_close_points_in_group(
    all_points_df: pd.DataFrame, duplicate_precision: float | None
) -> dict:
    """Select points that are closer to a given distance."""
    if len(all_points_df) <= 1:
        return {}

    tree = KDTree(all_points_df[TARGET_COORDS_COLS])
    close_pts = tree.query_pairs(duplicate_precision)

    if not close_pts:
        return {}

    # Find labels of duplicated points
    to_update: dict[Any, Any] = {}
    for a, b in close_pts:
        label_a = all_points_df.index[a]
        label_b = all_points_df.index[b]
        if label_a in to_update:
            to_update[label_a].add(label_b)
        elif label_b in to_update:
            to_update[label_b].add(label_a)
        else:
            to_update[label_a] = {label_b}

    return to_update


def drop_close_points(
    all_points_df: pd.DataFrame, duplicate_precision: float | None
) -> pd.DataFrame:
    """Drop points that are closer to a given distance."""
    if duplicate_precision is None or len(all_points_df) <= 1:
        return all_points_df

    all_groups = all_points_df.groupby(["morphology", "axon_id"], group_keys=True)
    deduplicated = all_groups[TARGET_COORDS_COLS].apply(
        lambda group: select_close_points_in_group(group, duplicate_precision)
    )

    # Format the labels
    def format_labels(to_update) -> list:
        skip = set()
        items = list(to_update.items())
        for num, (i, j) in enumerate(items):
            if i in skip:
                continue
            for ii, jj in items[num + 1 :]:
                if i in jj or ii in j:
                    j.update(jj)
                    skip.add(ii)
                    skip.update(jj)
        return [i for i in items if i[0] not in skip]

    to_update = list(itertools.chain.from_iterable(deduplicated.apply(format_labels)))

    # Update the terminal IDs
    for ref, changed in to_update:
        all_points_df.loc[list(changed), "terminal_id"] = all_points_df.loc[ref, "terminal_id"]

    if "level_2" in all_points_df.columns:
        all_points_df = all_points_df.drop(columns=["level_2"])

    return all_points_df


def pick_target_populations(probs, rng, logger=None, max_tries=10):
    """Pick target populations for each source point."""
    probs["random_number"] = pd.Series(-1, index=probs.index.to_numpy(), dtype=float)
    no_target_mask = probs["random_number"] < 0
    n_tries = 0
    mask_size = no_target_mask.sum()
    selected_mask = pd.Series(data=True, index=probs.index, dtype=bool)
    while n_tries < max_tries and mask_size > 0:
        # Select the populations according to the associated probabilities
        probs.loc[no_target_mask, "random_number"] = rng.uniform(size=mask_size)
        selected_mask = probs["random_number"] <= probs["target_probability"]

        # TODO: Here we implicitly suppose that we want to select at least 1 target per axon, but
        # maybe we want a customizable minimum number of targets?

        # Check which axons don't have any selected target
        no_target_mask = probs.merge(
            probs.loc[selected_mask, ["morphology", "axon_id", "random_number"]].drop_duplicates(
                subset=["morphology", "axon_id"]
            ),
            on=["morphology", "axon_id"],
            how="left",
            suffixes=("", "_tmp"),
        )["random_number_tmp"].isna()

        mask_size = no_target_mask.sum()
        n_tries = n_tries + 1

    if mask_size > 0 and logger is not None:
        logger.warning(
            "Could not find any target for the following morphologies: %s",
            ", ".join(
                [
                    f"{i[0]} (axon ID={i[1]})"
                    for i in probs.loc[no_target_mask, ["morphology", "axon_id"]]
                    .drop_duplicates()
                    .to_numpy()
                    .tolist()
                ]
            ),
        )

    return selected_mask


def get_target_points(
    source_points,
    target_probabilities,
    tufts_dist_df: pd.DataFrame | None = None,
    duplicate_precision: float | None = None,
    *,
    atlas: AtlasHelper | None = None,
    brain_regions_masks: File | None = None,
    rng: SeedType | None = None,
    max_tries: int = 10,
    output_path: FileType | None = None,
    logger: logging.Logger | logging.LoggerAdapter | None = None,
):
    """Find the target points for all given source points."""
    rng = np.random.default_rng(rng)
    if logger is None:
        logger = LOGGER

    # Create default populations if missing
    if "population_id" not in source_points.columns:
        source_points["population_id"] = DEFAULT_POPULATION
    if "source_population_id" not in target_probabilities.columns:
        target_probabilities["source_population_id"] = DEFAULT_POPULATION

    # Duplicated entries stand for different axons so we create axon IDs
    source_points["axon_id"] = source_points.groupby("morphology").cumcount()

    # Get ascendants in the hierarchy
    if atlas is not None and "st_level" not in source_points.columns:
        cells_region_parents = source_points.merge(
            atlas.brain_regions_and_ascendants,
            left_on="source_brain_region_id",
            right_on="id",
            how="left",
        ).drop(columns=["id"])
    else:
        cells_region_parents = source_points.copy(deep=False)
        cells_region_parents["st_level"] = 0

    # Remove useless columns before merge to reduce RAM usage
    cells_region_parents = cells_region_parents[
        ["morphology", "axon_id", "source_brain_region_id", "population_id", "st_level"]
    ]
    target_probabilities = target_probabilities[
        ["source_population_id", "target_brain_region_id", "target_population_id", "probability"]
        + (
            TARGET_COORDS_COLS
            if not set(TARGET_COORDS_COLS).difference(target_probabilities.columns)
            else []
        )
        + (["target_hemisphere"] if "target_hemisphere" in target_probabilities.columns else [])
    ]
    # Get the probabilities
    probs = cells_region_parents.merge(
        target_probabilities.rename(columns={"probability": "target_probability"}),
        left_on=["population_id"],
        right_on=["source_population_id"],
        how="left",
    )
    # Report missing probabilities
    missing_probs = probs.loc[probs["target_probability"].isna()]
    if len(missing_probs) > 0:
        logger.warning(
            "The following morphologies have no associated target probabilities: %s",
            missing_probs["morphology"].drop_duplicates().to_list(),
        )

    # Keep only the probabilities from the deepest level in the hierarchy
    probs = probs.dropna(axis=0, subset=["target_probability"])
    probs = probs.loc[
        probs["st_level"]
        == probs.groupby(["morphology", "source_brain_region_id"])["st_level"].transform("max")
    ].reset_index(drop=True)

    # Ensure that at least one region is selected for each morphology
    selected_mask = pick_target_populations(probs, rng, logger, max_tries)

    def draw_tuft_number(row) -> int:
        try:
            return max(
                1,
                int(
                    rng.normal(
                        tufts_dist_df.loc[row["target_population_id"], "mean_tuft_number"],  # type: ignore[union-attr]
                        tufts_dist_df.loc[row["target_population_id"], "std_tuft_number"],  # type: ignore[union-attr]
                    )
                ),
            )
        except (KeyError, AttributeError):
            return 1

    # Set the number of tufts to grow for each target population
    probs["num_tufts_to_grow"] = 0
    probs.loc[selected_mask, "num_tufts_to_grow"] = probs.apply(draw_tuft_number, axis=1)

    probs_cols = [
        "morphology",
        "axon_id",
        "source_brain_region_id",
        "target_population_id",
        "target_brain_region_id",
        "num_tufts_to_grow",
    ]
    if "target_hemisphere" in probs.columns:
        probs_cols.append("target_hemisphere")
    if not set(TARGET_COORDS_COLS).difference(probs.columns):
        probs_cols.extend(TARGET_COORDS_COLS)
    target_points = source_points.merge(
        probs.loc[
            selected_mask,
            probs_cols,
        ],
        on=["morphology", "axon_id", "source_brain_region_id"],
        how="left",
    ).dropna(subset=["target_population_id"])

    # Duplicate row according to the number of tuft in each population
    repeated_index = np.repeat(target_points.index, target_points["num_tufts_to_grow"].astype(int))

    # Reindex the DataFrame with the new repeated index
    target_points = target_points.loc[repeated_index].reset_index(drop=True)

    compute_coords(target_points, brain_regions_masks, atlas=atlas, rng=rng)
    # Build terminal IDs inside groups
    counter = target_points[["morphology", "axon_id"]].copy(deep=False)
    counter["counter"] = 1
    target_points["terminal_id"] = counter.groupby(["morphology", "axon_id"])["counter"].cumsum()

    other_columns = []
    if "seed" in target_points.columns:
        other_columns.append("seed")

    # Remove useless columns
    target_points = target_points[
        [
            "morphology",
            "morph_file",
            "axon_id",
            "terminal_id",
            *COORDS_COLS,
            "orientation",
            "grafting_section_id",
            "population_id",
            "source_brain_region_id",
            *SOURCE_COORDS_COLS,
            "target_population_id",
            "target_brain_region_id",
            *TARGET_COORDS_COLS,
            *other_columns,
        ]
    ].rename(
        columns={
            "population_id": "source_population_id",
        },
    )

    target_points = drop_close_points(target_points, duplicate_precision)

    # Export the target points
    if output_path is not None:
        with ignore_warnings(pd.errors.PerformanceWarning):
            target_points.to_hdf(output_path, key="target_points")
        logger.debug("Target points exported to %s", output_path)

    logger.info("Found %s target point(s)", len(target_points))

    return target_points.sort_values(["morphology", "axon_id", "terminal_id"]).reset_index(
        drop=True
    )



"""Store some constant values."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

from enum import IntEnum


class CoordsCols(list):
    """Class to associate column names to coordinates."""

    def __init__(self, *args):
        """Constructor of the CoordsCols class."""
        if len(args) != 3:
            msg = "Exactly 3 column names should be given"
            raise ValueError(msg)
        super().__init__(args)
        self.X = self[0]
        self.Y = self[1]
        self.Z = self[2]
        self.XYZ = [self.X, self.Y, self.Z]


# Point coordinates
COORDS_COLS = CoordsCols("x", "y", "z")
ATLAS_COORDS_COLS = CoordsCols("atlas_x", "atlas_y", "atlas_z")
COMMON_ANCESTOR_COORDS_COLS = CoordsCols(
    "common_ancestor_x", "common_ancestor_y", "common_ancestor_z"
)
SOURCE_COORDS_COLS = CoordsCols("source_x", "source_y", "source_z")
TARGET_COORDS_COLS = CoordsCols("target_x", "target_y", "target_z")
TUFT_COORDS_COLS = CoordsCols("tuft_x", "tuft_y", "tuft_z")

# Graph coordinates
FROM_COORDS_COLS = CoordsCols("x_from", "y_from", "z_from")
TO_COORDS_COLS = CoordsCols("x_to", "y_to", "z_to")

# Constants
AXON_GRAFTING_POINT_HDF_GROUP = "axon_grafting_points"
DEFAULT_OUTPUT_PATH = "out"
DEFAULT_POPULATION = "default"
WMR_ATLAS_ID = "id"  # Can also be "atlas_id" in other atlas versions

NodeProvider = IntEnum(
    "NodeProvider", ["source", "target", "intermediate", "random", "attractors", "bbox", "Voronoi"]
)



"""Create the source points from the atlas."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

import logging
from pathlib import Path

import numpy as np
import pandas as pd
from morphio import Morphology
from morphio import RawDataError
from morphio import SectionType
from neurom import COLS

from axon_synthesis.atlas import AtlasHelper
from axon_synthesis.constants import COORDS_COLS
from axon_synthesis.constants import DEFAULT_POPULATION
from axon_synthesis.constants import SOURCE_COORDS_COLS
from axon_synthesis.typing import FileType
from axon_synthesis.typing import SeedType
from axon_synthesis.utils import load_morphology

_MAX_DISPLAYED = 100


def section_id_to_position(morph, sec_id):
    """Find the position of the last point of the section from its ID."""
    morph = Morphology(morph)
    try:
        return morph.section(sec_id).points[-1, COLS.XYZ] - morph.soma.center
    except RawDataError:
        return None


def find_existing_axons(morph_file):
    """Find the positions of the existing axons in a morphology."""
    morph = load_morphology(morph_file, recenter=True)
    return [sec.points[0] for sec in morph.root_sections if sec.type == SectionType.axon]


def map_population(
    cells_df: pd.DataFrame,
    atlas: AtlasHelper | None,
    populations: pd.DataFrame | None = None,
    *,
    rng: SeedType = None,
):
    """Find the population given the position of the morphology and the populations."""
    if "population_id" in cells_df.columns:
        return cells_df

    rng = np.random.default_rng(rng)
    if populations is None:
        cells_df["population_id"] = DEFAULT_POPULATION
    elif atlas is None or populations is None:
        msg = (
            "The 'populations' and 'atlas' arguments should not be None when the 'population_id' "
            "column is not given"
        )
        raise RuntimeError(msg)
    else:
        # Get all the parent IDs in the brain region hierarchy
        cells_region_parents = cells_df.merge(
            atlas.brain_regions_and_ascendants,
            left_on="source_brain_region_id",
            right_on="id",
            how="left",
        )

        # Get the probabilities
        probs = cells_region_parents.merge(
            populations.rename(columns={"probability": "population_probability"}),
            left_on="elder_id",
            right_on="brain_region_id",
            how="left",
        )
        probs = probs.dropna(axis=0, subset=["population_id"])

        # Keep only the probabilities from the deepest level in the hierarchy
        probs = probs.loc[
            probs["st_level"]
            == probs.groupby(["morphology", "source_brain_region_id"])["st_level"].transform("max")
        ]

        # if probs has column 'hemisphere', select probabilities also on hemispheres
        if "hemisphere" in probs.columns:
            # rename 'hemisphere' to 'population_hemisphere' to not mistake with the cell
            probs = probs.rename(columns={"hemisphere": "population_hemisphere"})
            # add a column 'cell_hemisphere' for each cell
            probs["cell_hemisphere"] = probs[COORDS_COLS].apply(atlas.get_hemisphere, axis=1)
            # and drop the rows where the cell and the population are not in the same hemisphere
            probs = probs.loc[probs["cell_hemisphere"] == probs["population_hemisphere"]]

        # Select the populations according to the associated probabilities
        selected = probs.groupby(["morphology", "source_brain_region_id"]).sample(
            weights=probs["population_probability"],
            random_state=rng,
        )

        cells_df = cells_df.merge(
            selected[["morphology", "source_brain_region_id", "population_id"]],
            on=["morphology", "source_brain_region_id"],
            how="left",
        ).fillna({"population_id": DEFAULT_POPULATION})

    return cells_df


def fill_morph_file_col(cells_df, morph_dir):
    """Find the files corresponding to the given morphology names.

    .. warning::

        The validity of the morphology files is not tested but only their existence.
    """
    cells_df.loc[:, "morph_file"] = None
    for ext in [".asc", ".h5", ".swc"]:
        new_paths = (Path(morph_dir) / cells_df["morphology"]).apply(
            lambda x, ext=ext: x.with_suffix(ext).resolve()
        )
        existing_paths_mask = new_paths.apply(lambda x: x.exists())
        cells_df.loc[existing_paths_mask, "morph_file"] = new_paths.loc[existing_paths_mask]
    if cells_df["morph_file"].isna().any():
        msg = (
            f"Could not find morphology files in the '{morph_dir}' directory for the following "
            "morphologies: ["
        )
        missing_files = cells_df.loc[cells_df["morph_file"].isna(), "morphology"]
        nb_tot = len(missing_files)
        if nb_tot > _MAX_DISPLAYED:
            missing_files = missing_files.head(_MAX_DISPLAYED)
            suffix = f", ... ({nb_tot} morphologies in total but only {_MAX_DISPLAYED} displayed)]"
        else:
            suffix = "]"
        msg += ", ".join(("'" + missing_files + "'").to_list()) + suffix
        raise RuntimeError(msg)


def combine_existing_axons(cells_df):
    """Combine rebuilt existing axons with explicitly given axons."""
    existing_axons = (
        cells_df.groupby("morph_file")["morph_file"]
        .apply(lambda group: find_existing_axons(group.name))
        .apply(pd.Series)
        .stack()
        .rename("XYZ")  # type: ignore[call-overload]
    )
    existing_axons.index.rename("axon_id", level=1, inplace=True)  # type: ignore[call-arg]
    existing_axons = existing_axons.reset_index()
    existing_axons["grafting_section_id"] = -1
    if existing_axons.empty:
        existing_axons[SOURCE_COORDS_COLS] = pd.DataFrame(
            {col: pd.Series(dtype=float) for col in SOURCE_COORDS_COLS}
        )
    else:
        existing_axons[SOURCE_COORDS_COLS] = np.stack(
            existing_axons["XYZ"].to_numpy()  # type: ignore[arg-type]
        )
    new_axons = (
        cells_df[
            [
                col
                for col in cells_df.columns
                if col not in ["grafting_section_id", *SOURCE_COORDS_COLS]
            ]
        ]
        .drop_duplicates("morphology")
        .merge(
            existing_axons[["morph_file", "grafting_section_id", *SOURCE_COORDS_COLS]],
            on="morph_file",
            how="right",
        )
    )

    if "rebuilt_existing_axon_id" in cells_df.columns:
        if cells_df[["morphology", "rebuilt_existing_axon_id"]].duplicated().any():
            msg = "The 'rebuilt_existing_axon_id' should be unique for each morphology"
            raise ValueError(msg)
        new_axons["rebuilt_existing_axon_id"] = new_axons.groupby("morphology").cumcount()
        all_axons = (
            new_axons.reset_index()
            .merge(
                cells_df[["morphology", "rebuilt_existing_axon_id"]]
                .drop_duplicates()
                .reset_index(),
                on=["morphology", "rebuilt_existing_axon_id"],
                how="outer",
                indicator=True,
                suffixes=("_new_axons", "_cells_df"),
            )
            .dropna(subset=["morph_file"])
        )
        new_axons = new_axons.loc[
            all_axons.loc[all_axons["_merge"] == "left_only", "index_new_axons"]
            .astype(int)
            .to_numpy()
        ]
        cells_df = cells_df.loc[
            all_axons.loc[all_axons["_merge"] == "both", "index_cells_df"].astype(int).to_numpy()
        ]

    # We don't add axons starting from the soma when an existing axon is rebuilt
    cells_df = pd.concat([cells_df.loc[cells_df.index.difference(new_axons.index)], new_axons])
    cells_df.fillna(
        {source_col: 0 for source_col in SOURCE_COORDS_COLS},
        inplace=True,
    )
    return cells_df


def set_source_points(
    cells_df: pd.DataFrame,
    atlas: AtlasHelper | None,
    morph_dir: FileType,
    population_probabilities: pd.DataFrame | None = None,
    axon_grafting_points: pd.DataFrame | None = None,
    *,
    rng: SeedType = None,
    rebuild_existing_axons: bool = False,
    logger: logging.Logger | logging.LoggerAdapter | None = None,
):
    """Extract source points from a cell collection."""
    if "morph_file" not in cells_df.columns:
        fill_morph_file_col(cells_df, morph_dir)

    cells_df = cells_df.reset_index(drop=True)

    # Get source points from the axon_grafting_points file
    if axon_grafting_points is not None:
        axon_grafting_points = axon_grafting_points[
            [
                col
                for col in axon_grafting_points.columns
                if col
                in [
                    "morphology",
                    "grafting_section_id",
                    *SOURCE_COORDS_COLS,
                    "population_id",
                    "rebuilt_existing_axon_id",
                ]
            ]
        ]
        cells_df = cells_df.merge(axon_grafting_points, on="morphology", how="left")

    # Find existing axons to rebuild them if required
    if rebuild_existing_axons:
        cells_df = combine_existing_axons(cells_df)

    # Format the grafting_section_id column
    if "grafting_section_id" not in cells_df.columns:
        cells_df["grafting_section_id"] = -1
    else:
        cells_df["grafting_section_id"] = cells_df["grafting_section_id"].fillna(-1).astype(int)

    # If some coordinate columns are missing we reset them
    if len(set(SOURCE_COORDS_COLS).difference(cells_df.columns)) > 0:
        cells_df[SOURCE_COORDS_COLS] = np.nan

    # Find where the coordinates should be updated
    missing_coords_mask = cells_df[SOURCE_COORDS_COLS].isna().any(axis=1)
    section_id_mask = (cells_df["grafting_section_id"] != -1) & missing_coords_mask

    # If no section ID is provided we start the axon from the center of the morphology
    if missing_coords_mask.any():
        cells_df.loc[missing_coords_mask, SOURCE_COORDS_COLS] = 0

    # We shift all the coordinates to the positions in the atlas
    cells_df[SOURCE_COORDS_COLS] += cells_df[COORDS_COLS].to_numpy()

    # If a section ID is provided we start the axon from the last point of this section
    # Note: The coordinates of the points of each morphology are relative to the center of this
    # morphology
    if section_id_mask.any():
        cells_df.loc[section_id_mask, SOURCE_COORDS_COLS] += (
            cells_df.loc[section_id_mask]
            .apply(
                lambda row: section_id_to_position(row["morph_file"], row["grafting_section_id"]),
                axis=1,
            )
            .apply(pd.Series)
            .to_numpy()
        )

    # Set atlas regions
    if "source_brain_region_id" not in cells_df.columns:
        if atlas is None:
            msg = "The 'source_brain_region_id' column is missing and no atlas is provided"
            raise ValueError(msg)
        cells_df["source_brain_region_id"] = atlas.brain_regions.lookup(
            cells_df[COORDS_COLS].to_numpy()
        )

    # Choose population
    source_points = map_population(cells_df, atlas, population_probabilities, rng=rng)

    if logger is not None:
        logger.debug("Found %s source point(s)", len(source_points))
    return source_points



"""Compute the Steiner Tree.

The solution is computed using the package pcst_fast: https://github.com/fraenkel-lab/pcst_fast
"""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

import logging
from pathlib import Path

import pandas as pd
import pcst_fast as pf
import plotly.graph_objs as go

from axon_synthesis.constants import COORDS_COLS
from axon_synthesis.constants import FROM_COORDS_COLS
from axon_synthesis.constants import TO_COORDS_COLS
from axon_synthesis.typing import FileType
from axon_synthesis.utils import build_layout_properties
from axon_synthesis.utils import sublogger


def plot_solution(nodes, edges, figure_path, logger=None):
    """Plot the given triangulation for debugging purpose."""
    segments = edges.copy(deep=False)
    segments["cutter"] = None

    segments["width"] = segments["weight"] / segments["length"]

    fig = go.Figure()

    edges_trace = go.Scatter3d(
        x=segments.loc[~segments["is_solution"], [FROM_COORDS_COLS.X, TO_COORDS_COLS.X, "cutter"]]
        .to_numpy()
        .flatten(),
        y=segments.loc[~segments["is_solution"], [FROM_COORDS_COLS.Y, TO_COORDS_COLS.Y, "cutter"]]
        .to_numpy()
        .flatten(),
        z=segments.loc[~segments["is_solution"], [FROM_COORDS_COLS.Z, TO_COORDS_COLS.Z, "cutter"]]
        .to_numpy()
        .flatten(),
        line={"width": 1, "color": "black"},
        mode="lines",
        name="Steiner graph",
    )

    solution_trace = go.Scatter3d(
        x=segments.loc[segments["is_solution"], [FROM_COORDS_COLS.X, TO_COORDS_COLS.X, "cutter"]]
        .to_numpy()
        .flatten(),
        y=segments.loc[segments["is_solution"], [FROM_COORDS_COLS.Y, TO_COORDS_COLS.Y, "cutter"]]
        .to_numpy()
        .flatten(),
        z=segments.loc[segments["is_solution"], [FROM_COORDS_COLS.Z, TO_COORDS_COLS.Z, "cutter"]]
        .to_numpy()
        .flatten(),
        line={"width": 5, "color": "red"},
        mode="lines",
        name="Steiner solution",
    )

    edge_weights_trace = go.Scatter3d(
        x=0.5 * (segments[FROM_COORDS_COLS.X] + segments[TO_COORDS_COLS.X]).to_numpy(),
        y=0.5 * (segments[FROM_COORDS_COLS.Y] + segments[TO_COORDS_COLS.Y]).to_numpy(),
        z=0.5 * (segments[FROM_COORDS_COLS.Z] + segments[TO_COORDS_COLS.Z]).to_numpy(),
        mode="markers",
        opacity=0,
        showlegend=False,
        hovertemplate="Weight of edge %{hovertext}<extra></extra>",
        hovertext=[
            f"{idx} = {val['weight']} (length ratio = {val['width']})"
            for idx, val in segments[["weight", "width"]].round(2).to_dict("index").items()
        ],
        marker={"size": 4},
    )

    source_point_trace = go.Scatter3d(
        x=[nodes.loc[0, COORDS_COLS.X]],
        y=[nodes.loc[0, COORDS_COLS.Y]],
        z=[nodes.loc[0, COORDS_COLS.Z]],
        marker={"color": "purple", "size": 4},
        mode="markers",
        name="Source point",
    )

    intermediate_points_trace = go.Scatter3d(
        x=nodes.loc[~nodes["is_terminal"], COORDS_COLS.X].to_numpy(),
        y=nodes.loc[~nodes["is_terminal"], COORDS_COLS.Y].to_numpy(),
        z=nodes.loc[~nodes["is_terminal"], COORDS_COLS.Z].to_numpy(),
        marker={"color": "green", "size": 2},
        mode="markers",
        name="Intermediate points",
    )

    target_points_trace = go.Scatter3d(
        x=nodes.loc[nodes["is_terminal"], COORDS_COLS.X].to_numpy()[1:],
        y=nodes.loc[nodes["is_terminal"], COORDS_COLS.Y].to_numpy()[1:],
        z=nodes.loc[nodes["is_terminal"], COORDS_COLS.Z].to_numpy()[1:],
        marker={"color": "chocolate", "size": 4},
        mode="markers",
        name="Target points",
    )

    fig.add_trace(edges_trace)
    fig.add_trace(solution_trace)
    fig.add_trace(edge_weights_trace)
    fig.add_trace(intermediate_points_trace)
    fig.add_trace(source_point_trace)
    fig.add_trace(target_points_trace)

    layout_props = build_layout_properties(nodes[COORDS_COLS].to_numpy(), 0.1)

    fig.update_scenes(layout_props)
    fig.update_layout(title=Path(figure_path).stem)

    # Export figure
    Path(figure_path).parent.mkdir(parents=True, exist_ok=True)
    fig.write_html(figure_path)

    if logger is not None:
        logger.debug("Exported Steiner Tree solution figure to %s", figure_path)


def compute_solution(
    nodes: pd.DataFrame,
    edges: pd.DataFrame,
    *,
    output_path_nodes: FileType | None = None,
    output_path_edges: FileType | None = None,
    figure_path: FileType | None = None,
    logger: logging.Logger | logging.LoggerAdapter | None = None,
):
    """Compute the Steiner Tree solution from the given nodes and edges."""
    logger = sublogger(logger, __name__)

    nodes["is_solution"] = False
    edges["is_solution"] = False

    logger.debug(
        "%s nodes and %s edges",
        len(nodes),
        len(edges),
    )

    # Prepare prizes: we want to connect all terminals so we give them an 'infinite' prize
    prizes = 100.0 * nodes["is_terminal"] * edges["weight"].sum()

    # Compute Steiner Tree
    solution_nodes, solution_edges = pf.pcst_fast(  # pylint: disable=c-extension-no-member
        edges[["from", "to"]].to_numpy(),
        prizes,
        edges["weight"].to_numpy(),
        -1,
        1,
        "gw",
        0,
    )

    nodes.loc[
        (nodes["id"].isin(solution_nodes)),
        "is_solution",
    ] = True

    group_edge_ids = edges.reset_index()["index"]
    edge_ids = pd.Series(-1, index=edges.index)
    reverted_group_edge_ids = pd.Series(
        group_edge_ids.index.to_numpy(), index=group_edge_ids.to_numpy()
    )
    edge_ids.loc[reverted_group_edge_ids.index] = reverted_group_edge_ids
    edges.loc[
        (edge_ids.isin(solution_edges)),
        "is_solution",
    ] = True

    if output_path_nodes is not None:
        # Export the solution nodes
        nodes.to_feather(str(output_path_nodes))
    if output_path_edges is not None:
        # Export the solution edges
        edges.to_feather(str(output_path_edges))

    in_solution_nodes = nodes.loc[nodes["is_solution"]]
    in_solution_edges = edges.loc[edges["is_solution"]]

    logger.debug(
        "The solution contains %s among %s nodes and %s among %s edges",
        len(in_solution_nodes),
        len(nodes),
        len(in_solution_edges),
        len(edges),
    )

    # Add node data to solution edges
    in_solution_edges = in_solution_edges.merge(
        in_solution_nodes[["terminal_id", "is_terminal"]].rename(
            columns={"terminal_id": "source_terminal_id", "is_terminal": "source_is_terminal"}
        ),
        left_on="from",
        right_index=True,
        how="left",
    )
    in_solution_edges = in_solution_edges.merge(
        in_solution_nodes[["terminal_id", "is_terminal"]].rename(
            columns={"terminal_id": "target_terminal_id", "is_terminal": "target_is_terminal"}
        ),
        left_on="to",
        right_index=True,
        how="left",
    )

    if figure_path is not None:
        plot_solution(
            nodes,
            edges,
            figure_path,
            logger=logger,
        )

    return in_solution_nodes, in_solution_edges



"""Generate atlas for tests."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

# pylint: disable=missing-function-docstring
import json
import os
import shutil
from pathlib import Path
from subprocess import call

from voxcell.nexus.voxelbrain import Atlas

from . import DATA


def generate_small_O1(atlas_dir):  # noqa: N802
    """Dump a small O1 atlas in folder path."""
    atlas_dir = Path(atlas_dir)

    # fmt: off
    with Path(os.devnull).open("w", encoding="utf-8") as f:
        call(["brainbuilder", "atlases",  # noqa: S603, S607
              "-n", "6,5,4,3,2,1",
              "-t", "200,100,100,100,100,200",
              "-d", "100",
              "-o", str(atlas_dir),
              "column",
              "-a", "1000",
              ], stdout=f, stderr=f)
    # fmt: on

    # Add metadata and region structure
    shutil.copyfile(DATA / "atlas_metadata.json", atlas_dir / "metadata.json")
    shutil.copyfile(DATA / "region_structure.yaml", atlas_dir / "region_structure.yaml")

    # Split each brain region into 2 sub-regions
    atlas = Atlas.open(str(atlas_dir))
    brain_regions = atlas.load_data("brain_regions")
    brain_regions.raw[11:, :, :] *= 10
    brain_regions.save_nrrd(str(atlas_dir / "brain_regions.nrrd"))

    # Update the brain region hierarchy
    with (atlas_dir / "hierarchy.json").open(encoding="utf-8") as f:
        region_map = json.load(f)
    region_map["children"][0]["acronym"] = "mc0"
    region_map["children"].append(
        {
            "id": 10,
            "acronym": "mc1",
            "name": "hypercolumn 1",
            "children": [
                {"id": 20, "acronym": "mc1;6", "name": "hypercolumn 1, 6"},
                {"id": 30, "acronym": "mc1;5", "name": "hypercolumn 1, 5"},
                {"id": 40, "acronym": "mc1;4", "name": "hypercolumn 1, 4"},
                {"id": 50, "acronym": "mc1;3", "name": "hypercolumn 1, 3"},
                {"id": 60, "acronym": "mc1;2", "name": "hypercolumn 1, 2"},
                {"id": 70, "acronym": "mc1;1", "name": "hypercolumn 1, 1"},
            ],
        },
    )

    def recursive_atlas_id(data) -> None:
        if "id" in data:
            data["atlas_id"] = data["id"]
        if "children" in data:
            for i in data["children"]:
                recursive_atlas_id(i)

    recursive_atlas_id(region_map)

    with (atlas_dir / "hierarchy.json").open("w", encoding="utf-8") as f:
        json.dump(region_map, f, indent=4, sort_keys=True)

    return atlas_dir



"""Define the class to store the synthesis outputs."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

from pathlib import Path
from typing import ClassVar

from attrs import asdict
from attrs import define
from attrs import field

from axon_synthesis.base_path_builder import BasePathBuilder


@define
class OutputConfig:
    """Class to store the parameters for the outputs.

    Attributes:
        path: The path to the output directory.
        final_figures: Enable export of the final figures.
        graph_creation_figures: Enable export of the graph creation figures.
        graph_creation_data: Enable export of the graph creation data.
        main_trunk_figures: Enable export of the main trunk figures.
        main_trunk_morphologies: Enable export of the main trunk morphologies.
        morphologies: Enable export of the final morphologies.
        morphologies_edges: Enable export of the final morphologies as edges.
        postprocess_trunk_figures: Enable export of the postprocess trunk figures.
        postprocess_trunk_morphologies: Enable export of the postprocess trunk morphologies.
        steiner_tree_solution_figures: Enable export of the steiner tree solution figures.
        steiner_tree_solutions: Enable export of the steiner tree solutions.
        target_point_figures: Enable export of the target point figures.
        target_points: Enable export of the target points.
        tuft_figures: Enable export of the tuft figures.
        tuft_morphologies: Enable export of the tuft morphologies.
    """

    path: Path = field(converter=Path)
    final_figures: bool = False
    graph_creation_figures: bool = False
    graph_creation_data: bool = False
    main_trunk_figures: bool = False
    main_trunk_morphologies: bool = False
    morphologies: bool = True
    morphologies_edges: bool = False
    postprocess_trunk_figures: bool = False
    postprocess_trunk_morphologies: bool = False
    steiner_tree_solution_figures: bool = False
    steiner_tree_solutions: bool = False
    target_point_figures: bool = False
    target_points: bool = False
    tuft_figures: bool = False
    tuft_morphologies: bool = False


class Outputs(BasePathBuilder):
    """Class to store the synthesis outputs."""

    def __init__(self, config: OutputConfig, *, exists: bool = False, create: bool = False):
        """Constructor for the Outputs class."""
        super().__init__(config.path, exists=exists, create=create)
        self.config = config

        for name in asdict(self.config):  # pylint: disable=not-an-iterable
            if name == "path":
                continue
            if getattr(self.config, name, False):
                self._optional_keys.discard(name.upper())
            else:
                self._filenames[name.upper()] = None
        self._reset_attributes()

    _filenames: ClassVar[dict] = {
        "FINAL_FIGURES": "FinalFigures",
        "GRAPH_CREATION_FIGURES": "GraphCreationFigures",
        "GRAPH_CREATION_DATA": "GraphCreationData",
        "MAIN_TRUNK_FIGURES": "MainTrunkFigures",
        "MAIN_TRUNK_MORPHOLOGIES": "MainTrunkMorphologies",
        "MORPHOLOGIES": "Morphologies",
        "MORPHOLOGIES_EDGES": "MorphologiesEdges",
        "POSTPROCESS_TRUNK_FIGURES": "PostProcessTrunkFigures",
        "POSTPROCESS_TRUNK_MORPHOLOGIES": "PostProcessTrunkMorphologies",
        "STEINER_TREE_SOLUTION_FIGURES": "SteinerTreeSolutionFigures",
        "STEINER_TREE_SOLUTIONS": "SteinerTreeSolutions",
        "TARGET_POINT_FIGURES": "TargetPointsFigures",
        "TARGET_POINTS": "target_points.h5",
        "TUFT_FIGURES": "TuftFigures",
        "TUFT_MORPHOLOGIES": "TuftMorphologies",
    }

    _optional_keys: ClassVar[set[str]] = {
        "FINAL_FIGURES",
        "GRAPH_CREATION_FIGURES",
        "GRAPH_CREATION_DATA",
        "MAIN_TRUNK_FIGURES",
        "MAIN_TRUNK_MORPHOLOGIES",
        "MORPHOLOGIES_EDGES",
        "POSTPROCESS_TRUNK_FIGURES",
        "POSTPROCESS_TRUNK_MORPHOLOGIES",
        "STEINER_TREE_SOLUTION_FIGURES",
        "STEINER_TREE_SOLUTIONS",
        "TARGET_POINT_FIGURES",
        "TARGET_POINTS",
        "TUFT_FIGURES",
        "TUFT_MORPHOLOGIES",
    }

    _dir_keys: ClassVar[set[str]] = {
        "FINAL_FIGURES",
        "GRAPH_CREATION_FIGURES",
        "GRAPH_CREATION_DATA",
        "MAIN_TRUNK_FIGURES",
        "MAIN_TRUNK_MORPHOLOGIES",
        "MORPHOLOGIES",
        "MORPHOLOGIES_EDGES",
        "POSTPROCESS_TRUNK_FIGURES",
        "POSTPROCESS_TRUNK_MORPHOLOGIES",
        "STEINER_TREE_SOLUTION_FIGURES",
        "STEINER_TREE_SOLUTIONS",
        "TARGET_POINT_FIGURES",
        "TUFT_FIGURES",
        "TUFT_MORPHOLOGIES",
    }



"""Post-process the Steiner solutions."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

import logging
from collections.abc import Sequence
from itertools import chain
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
import numpy.typing as npt
import pandas as pd
from attrs import define
from attrs import field
from attrs import validators
from neurom import morphmath
from neurom.apps import morph_stats
from neurom.core import Morphology
from neurom.core import Neurite
from neurots.morphmath import rotation
from plotly.subplots import make_subplots
from plotly_helper.neuron_viewer import NeuronBuilder

from axon_synthesis.typing import FileType
from axon_synthesis.typing import SeedType
from axon_synthesis.utils import add_camera_sync
from axon_synthesis.utils import build_layout_properties
from axon_synthesis.utils import save_morphology
from axon_synthesis.utils import sublogger

if TYPE_CHECKING:
    from morphio.mut import Section

WEIGHT_DISTANCE_TOLERANCE = 1e-8

HistoryType = tuple[list[float], (list[list[float] | npt.NDArray[np.floating]])]


@define
class PostProcessConfig:
    """Class to store the parameters needed for long-range trunk post-processing.

    Attributes:
        skip: Skip the post-processing step if set to True.
        history_path_length: The length used to compute the random walk history.
        default_history_path_length_coeff: The coefficient used to compute the history path length
            when it is not provided.
        global_target_coeff: The coefficient applied to the global target term.
        global_target_sigma_coeff: The sigma coefficient applied to the global target term.
        target_coeff: The coefficient applied to the next target term.
        target_sigma_coeff: The sigma coefficient applied to the next target term.
        random_coeff: The coefficient applied to the random term.
        history_coeff: The coefficient applied to the history term.
        history_sigma_coeff: The sigma coefficient applied to the history term.
        length_coeff: The coefficient applied to step length.
        max_random_direction_picks: The maximum number of random direction picks (random directions
            are picked again when they are not facing the target)
    """

    skip: bool = False
    history_path_length: float | None = field(
        default=None, validator=validators.optional(validators.ge(0))
    )
    default_history_path_length_coeff: float = field(default=5, validator=validators.gt(0))
    global_target_coeff: float = field(default=0, validator=validators.ge(0))
    global_target_sigma_coeff: float = field(default=10, validator=validators.gt(0))
    target_coeff: float = field(default=2, validator=validators.ge(0))
    target_sigma_coeff: float = field(default=2, validator=validators.gt(0))
    random_coeff: float = field(default=2, validator=validators.ge(0))
    history_coeff: float = field(default=2, validator=validators.ge(0))
    history_sigma_coeff: float = field(default=2, validator=validators.gt(0))
    length_coeff: float = field(default=1, validator=validators.gt(0))
    max_random_direction_picks: int = field(default=10, validator=validators.ge(1))


def get_random_vector(
    distance: float = 1.0,
    norm: float | None = None,
    std: float | None = None,
    initial_theta: float | None = None,
    initial_phi: float | None = None,
    rng=np.random,
) -> npt.NDArray[np.floating]:
    """Return 3-d coordinates of a new random point.

    The distance between the produced point and (0,0,0) is given by the 'distance' argument.
    """
    # pylint: disable=assignment-from-no-return
    phi = rng.uniform(0.0, 2.0 * np.pi)
    theta = (
        rng.normal(norm, std)
        if norm is not None and std is not None
        else np.arccos(rng.uniform(-1.0, 1.0))
    )

    if initial_theta:
        theta += initial_theta
    if initial_phi:
        phi += initial_phi

    sn_theta = np.sin(theta)

    x = distance * np.cos(phi) * sn_theta
    y = distance * np.sin(phi) * sn_theta
    z = distance * np.cos(theta)

    return np.array((x, y, z))


def weights(lengths, history_path_length):
    """Compute the weights depending on the lengths."""
    return np.exp(np.append(np.cumsum(lengths[:-1]) - history_path_length + lengths[0], 0))


def history(latest_lengths, latest_directions, history_path_length):
    """Returns a combination of the segments history."""
    if len(latest_directions) == 0:
        return np.zeros(3)
    weighted_history = np.dot(weights(latest_lengths, history_path_length), latest_directions)

    distance = np.linalg.norm(weighted_history)
    if distance > WEIGHT_DISTANCE_TOLERANCE:
        weighted_history /= distance

    return weighted_history


def closest_seg_pt(pt, seg):
    """Compute the closest point on a line from a given point."""
    closest_pt = None
    u = seg[0] - pt
    v = seg[1] - seg[0]
    t = -np.dot(v, u) / np.dot(v, v)
    closest_pt = (1 - t) * seg[0] + t * seg[1]
    return closest_pt, t


def compute_direction(start_pt, end_pt):
    """Compute the direction and distance between 2 points."""
    direction = end_pt - start_pt
    distance = np.linalg.norm(direction)
    direction /= distance
    return direction, distance


def compute_step_direction(  # noqa: PLR0913
    intermediate_pts,
    current_pt,
    target,
    target_index,
    next_target,
    next_target_index,
    step_length,
    total_length,
    config,
    history_direction,
    latest_directions,
    rng=np.random,
    logger=None,
):
    """Compute the direction for one step."""
    # Compute the direction to the last target
    global_target_direction, global_target_dist = compute_direction(
        current_pt, intermediate_pts[-1]
    )

    # TODO: Should check that there is no other possible target between the last target and
    # this target that is further to this target.
    target_vec = target - current_pt
    target_dist = np.linalg.norm(target_vec)

    next_target_vec = next_target - current_pt

    current_target_coeff = np.exp(-target_dist / step_length)

    target_direction = (
        1 - current_target_coeff
    ) * target_vec + current_target_coeff * next_target_vec

    target_direction /= np.linalg.norm(target_direction)

    step_global_target_coeff = config.global_target_coeff * max(
        0,
        1
        + np.exp(
            -global_target_dist / (config.global_target_sigma_coeff * step_length)
        ),  # More when closer
    )
    step_target_coeff = config.target_coeff * max(
        0,
        1
        + np.exp(
            -target_dist / (config.target_sigma_coeff * step_length)
        )  # More near targets to pass closer
        - np.exp(
            -total_length / (config.target_sigma_coeff * step_length)
        ),  # Less at the beginning
    )
    step_history_coeff = config.history_coeff * max(
        0,
        1
        - np.exp(
            -target_dist / (config.history_sigma_coeff * step_length)
        )  # Less near targets to pass closer
        + np.exp(
            -total_length / (config.history_sigma_coeff * step_length)
        ),  # More at the beginning
    )
    step_random_coeff = config.random_coeff

    direction = -target_direction
    nb_rand = 0
    non_random_direction = (
        step_global_target_coeff * global_target_direction
        + step_target_coeff * target_direction
        + step_history_coeff * history_direction
    ).astype(float)

    # If the non random part of the direction does not head to the target direction
    # (e.g. because of the history), then we don't care if the resulting direction
    # does not head to the target direction
    heading_target = np.dot(target_direction, non_random_direction) >= 0

    while np.dot(direction, target_direction) < 0 and nb_rand < config.max_random_direction_picks:
        if nb_rand > 0:
            step_random_coeff = step_random_coeff / 2.0
        random_direction = get_random_vector(rng=rng)

        direction = (non_random_direction + step_random_coeff * random_direction).astype(float)

        if nb_rand == 0 and not heading_target:
            break

        nb_rand += 1

    direction /= np.linalg.norm(direction)

    if logger is not None:
        initial_phi, initial_theta = rotation.spherical_from_vector(latest_directions[-1])
        actual_target_direction = target_vec / np.linalg.norm(target_vec)
        composite_target_dist = (
            1 - current_target_coeff
        ) * target_dist + current_target_coeff * np.linalg.norm(next_target_vec)
        logger.debug(
            (
                "In random walk:\n\t"
                "global_target_dist=%s\n\t"
                "global_target_direction=%s\n\t"
                "total_length=%s\n\t"
                "step_length=%s\n\t"
                "target_index=%s\n\t"
                "target=%s\n\t"
                "target_vec=%s\n\t"
                "target_dist=%s\n\t"
                "target_direction=%s\n\t"
                "current_target_coeff=%s\n\t"
                "actual_target_direction=%s\n\t"
                "composite_target_dist=%s\n\t"
                "next_target=%s\n\t"
                "next_target_index=%s\n\t"
                "next_target_vec=%s\n\t"
                "random_direction=%s\n\t"
                "(initial_phi, initial_theta)=%s\n\t"
                "history_direction=%s\n\t"
                "step_global_target_coeff=%s\n\t"
                "step_target_coeff=%s\n\t"
                "step_random_coeff=%s\n\t"
                "step_history_coeff=%s\n\t"
                "direction=%s\n\t"
                "(phi, theta)=%s\n\t"
                "diff_direction=%s\n\t"
                "diff_actual_target_direction=%s\n\t"
                "diff_last_direction=%s\n\t"
                "current_pt=%s\n\t"
            ),
            global_target_dist,
            rotation.spherical_from_vector(global_target_direction),
            total_length,
            step_length,
            target_index,
            target,
            target_vec,
            target_dist,
            rotation.spherical_from_vector(target_direction),
            current_target_coeff,
            rotation.spherical_from_vector(actual_target_direction),
            composite_target_dist,
            next_target,
            next_target_index,
            next_target_vec,
            rotation.spherical_from_vector(random_direction),
            (initial_phi, initial_theta),
            rotation.spherical_from_vector(history_direction),
            step_global_target_coeff,
            step_target_coeff,
            step_random_coeff,
            step_history_coeff,
            direction,
            rotation.spherical_from_vector(direction),
            morphmath.angle_between_vectors(direction, target_direction),
            morphmath.angle_between_vectors(direction, actual_target_direction),
            morphmath.angle_between_vectors(direction, latest_directions[-1]),
            current_pt,
        )

    return direction, target_dist, global_target_dist


def check_next_target(  # noqa: PLR0913
    intermediate_pts,
    nb_intermediate_pts,
    target_index,
    target,
    target_dist,
    next_target_index,
    next_target,
    min_target_dist,
    length_norm,
    default_min_target_dist,
    logger,
):
    """Check if a next target can be selected and return it."""
    if target_dist >= min_target_dist:
        logger.debug("The random walk is going away from the target")
        new_target = True
    elif target_dist <= length_norm:
        logger.debug("The random walk reached the target %s", target_index)
        new_target = True
    else:
        new_target = False

    if new_target:
        logger.debug(
            "Changing target from %s to %s",
            target_index,
            next_target_index,
        )
        target_index = min(nb_intermediate_pts - 1, target_index + 1)
        target = intermediate_pts[target_index]
        next_target_index = min(nb_intermediate_pts - 1, target_index + 1)
        next_target = intermediate_pts[next_target_index]
        min_target_dist = default_min_target_dist

    return new_target, target_index, target, next_target_index, next_target, min_target_dist


def random_walk(
    starting_pt: Sequence[float] | npt.NDArray[np.floating],
    intermediate_pts: Sequence[Sequence[float]] | npt.NDArray[np.floating],
    length_stats: dict[str, float],
    # angle_stats,
    config: PostProcessConfig,
    previous_history: HistoryType | None = None,
    *,
    rng=np.random,
    logger: logging.Logger | logging.LoggerAdapter | None = None,
) -> tuple[npt.NDArray[np.floating], HistoryType]:
    """Perform a random walk guided by intermediate points."""
    logger = sublogger(logger, __name__)

    debug = logger.getEffectiveLevel() <= logging.DEBUG

    length_norm = length_stats["norm"] * config.length_coeff
    length_std = length_stats["std"] * config.length_coeff
    # angle_norm = angle_stats["norm"]
    # angle_std = angle_stats["std"]

    history_path_length = config.history_path_length
    if history_path_length is None:
        history_path_length = config.default_history_path_length_coeff * length_norm

    current_pt = np.array(starting_pt, dtype=float)
    new_intermediate_pts = np.array(intermediate_pts, dtype=float)
    new_pts: list[npt.NDArray[np.floating]] = [current_pt]

    total_length = 0

    # Compute the direction to the last target
    global_target_direction, global_target_dist = compute_direction(
        current_pt, new_intermediate_pts[-1]
    )

    # Compute the direction to the first target
    target_direction, target_dist = compute_direction(current_pt, new_intermediate_pts[0])

    # Setup initial history
    if previous_history:
        latest_lengths, latest_directions = previous_history
    else:
        nb_hist = int(history_path_length // length_norm)
        latest_lengths = [length_norm] * nb_hist
        latest_directions = [target_direction] * nb_hist

    nb_intermediate_pts = len(new_intermediate_pts)

    min_target_dist = global_target_dist * 2

    if debug:
        logger.debug(
            (
                "In random walk:\n\t"
                "global_target_dist=%s\n\t"
                "global_target_direction=%s\n\t"
                "target_direction=%s\n\t"
                "current_pt=%s\n\t"
                "intermediate_pts=%s\n\t"
            ),
            global_target_dist,
            global_target_direction,
            target_direction,
            current_pt,
            new_intermediate_pts,
        )

    target_index = 0
    target = new_intermediate_pts[target_index]
    next_target_index = min(nb_intermediate_pts - 1, 1)
    next_target = new_intermediate_pts[next_target_index]

    while global_target_dist >= length_norm:
        step_length = rng.normal(length_norm, length_std)
        while step_length <= 0:
            step_length = rng.normal(length_norm, length_std)

        history_direction = history(latest_lengths, latest_directions, history_path_length)

        direction, target_dist, global_target_dist = compute_step_direction(
            new_intermediate_pts,
            current_pt,
            target,
            target_index,
            next_target,
            next_target_index,
            step_length,
            total_length,
            config,
            history_direction,
            latest_directions,
            rng=rng,
            # logger=logger if debug else None,
        )

        (
            new_target,
            target_index,
            target,
            next_target_index,
            next_target,
            min_target_dist,
        ) = check_next_target(
            new_intermediate_pts,
            nb_intermediate_pts,
            target_index,
            target,
            target_dist,
            next_target_index,
            next_target,
            min_target_dist,
            length_norm,
            global_target_dist * 2,
            logger=logger,
        )

        if new_target:
            continue
        min_target_dist = min(min_target_dist, target_dist)

        current_pt = current_pt + direction * step_length
        total_length += step_length
        new_pts.append(current_pt)
        latest_lengths.append(step_length)
        latest_directions.append(direction)
        while sum(latest_lengths) >= history_path_length:
            latest_lengths.pop(0)
            latest_directions.pop(0)

    new_pts.append(new_intermediate_pts[-1])

    return np.array(new_pts, dtype=float), (latest_lengths, latest_directions)


def plot(morph, initial_morph, figure_path):
    """Plot the morphology after post-processing."""
    morph_name = figure_path.stem

    steiner_builder = NeuronBuilder(
        initial_morph,
        "3d",
        line_width=4,
        title=f"{morph_name}",
    )
    fig_builder = NeuronBuilder(morph, "3d", line_width=4, title=f"{morph_name}")

    fig = make_subplots(
        cols=2,
        specs=[[{"type": "scene"}, {"type": "scene"}]],
        subplot_titles=["Post-processed morphology", "Raw Steiner morphology"],
    )
    current_data = fig_builder.get_figure()["data"]
    steiner_data = steiner_builder.get_figure()["data"]
    all_data = list(chain(current_data, steiner_data))
    fig.add_traces(
        all_data,
        rows=[1] * (len(current_data) + len(steiner_data)),
        cols=[1] * len(current_data) + [2] * len(steiner_data),
    )

    layout_props = build_layout_properties(morph.points, 0.1)

    fig.update_scenes(layout_props)
    fig.update_layout(title=morph_name)

    # Export figure
    Path(figure_path).parent.mkdir(parents=True, exist_ok=True)
    fig.write_html(figure_path)

    # Update the HTML file to synchronize the cameras between the two plots
    add_camera_sync(figure_path)


def gather_sections(root_section, tuft_barcodes):
    """Gather the sections with unifurcations."""
    sections_to_smooth: list[list[Section]] = [[]]
    sec_use_parent = {
        tuple(i) for i in tuft_barcodes[["section_id", "use_parent"]].to_numpy().tolist()
    }
    for section in root_section.iter():
        sections_to_smooth[-1].append(section)
        if (
            len(section.children) != 1
            or (section.id, False) in sec_use_parent
            or any((child.id, True) in sec_use_parent for child in section.children)
        ):
            sections_to_smooth.append([])

    return [i for i in sections_to_smooth if i]


def resample_diameters(pts, resampled_pts, diams):
    """Resample the diameters on the new points."""
    path_lengths = np.insert(
        np.cumsum(np.linalg.norm(pts[1:] - pts[:-1], axis=1)),
        0,
        0,
    )
    new_path_lengths = np.insert(
        np.cumsum(np.linalg.norm(resampled_pts[1:] - resampled_pts[:-1], axis=1)),
        0,
        0,
    )
    return np.interp(
        new_path_lengths / new_path_lengths[-1],
        path_lengths / path_lengths[-1],
        diams,
    )


def export(morph, initial_morph, output_path, figure_path, logger):
    """Export morphology and figure."""
    # Export the new morphology
    if output_path is not None:
        save_morphology(morph, output_path, msg=f"Export morphology to {output_path}")

    # Create a figure of the new morphology
    if figure_path is not None:
        logger.info("Export figure to %s", figure_path)
        plot(morph, initial_morph, figure_path)


def post_process_trunk(
    morph: Morphology,
    trunk_section_id: int,
    trunk_properties: pd.DataFrame,
    tuft_barcodes: pd.DataFrame,
    config: PostProcessConfig,
    *,
    rng: SeedType = None,
    output_path: FileType | None = None,
    figure_path: FileType | None = None,
    logger: logging.Logger | logging.LoggerAdapter | None = None,
):
    """Post-process a trunk of the given morphology."""
    logger = sublogger(logger, __name__)
    debug = logger.getEffectiveLevel() <= logging.DEBUG

    if config.skip:
        logger.info("Skip post-processing")

    rng = np.random.default_rng(rng)

    logger.debug("Use the following post-processing config: %s", config)

    initial_morph = Morphology(morph) if figure_path is not None else None

    root_section = morph.section(trunk_section_id)

    # Get some statistics
    # TODO: Pick properties in a less random way? Maybe we could use the source region ID?
    ref_trunk_props = trunk_properties.sample(random_state=rng).iloc[0]

    if debug:
        logger.debug(
            "Ref statistics of the trunk: %s",
            ref_trunk_props.drop(
                ["morph_file", "axon_id"]
                + [row for row in ref_trunk_props.index if row.startswith("raw_")],
            ).to_dict(),
        )
        trunk_stats = morph_stats.extract_stats(
            Neurite(root_section),
            {
                "neurite": {
                    "segment_lengths": {"modes": ["mean", "std"]},
                    "segment_meander_angles": {"modes": ["mean", "std"]},
                },
            },
        )["axon"]
        logger.debug("Current trunk statistics: %s", trunk_stats)

    # Gather sections with unifurcations into future sections
    sections_to_smooth = gather_sections(root_section, tuft_barcodes)

    length_stats = {
        "norm": ref_trunk_props["mean_segment_lengths"],
        "std": ref_trunk_props["std_segment_lengths"],
    }
    # angle_stats = {
    #     "norm": ref_trunk_props["mean_segment_meander_angles"],
    #     "std": ref_trunk_props["std_segment_meander_angles"],
    # }

    # Smooth the sections but do not move the tuft roots
    parent_histories: dict[int, HistoryType] = {}
    for i in sections_to_smooth:
        pts = np.concatenate([i[0].points] + [sec.points[1:] for sec in i[1:]])
        diams = np.concatenate([i[0].diameters] + [sec.diameters[1:] for sec in i[1:]])
        min_diam_not_zero = np.min(diams, initial=1, where=diams != 0)
        diams = np.clip(diams, a_min=min_diam_not_zero, a_max=np.inf)

        if not i[0].is_root:
            try:
                parent_history = parent_histories[i[0].parent]
            except KeyError:
                parent_history = None
        else:
            parent_history = None

        resampled_pts, last_history = random_walk(
            pts[0],
            pts[1:],
            length_stats,
            # angle_stats,
            config,
            parent_history,
            rng=rng,
            logger=logger,
        )
        parent_histories[i[-1]] = last_history

        if len(resampled_pts) <= len(pts):
            resampled_pts = pts

        resampled_diams = resample_diameters(pts, resampled_pts, diams)

        # Update section points and diameters
        sec_pts = np.array_split(resampled_pts, len(i))
        sec_diams = np.array_split(resampled_diams, len(i))
        for num, sec in enumerate(i):
            if num == 0:
                s_pts = sec_pts[num]
                s_diams = sec_diams[num]
            else:
                s_pts = np.concatenate([[sec_pts[num - 1][-1]], sec_pts[num]])
                s_diams = np.concatenate([[sec_diams[num - 1][-1]], sec_diams[num]])
            sec.points = s_pts
            sec.diameters = s_diams

    if debug:
        trunk_stats = morph_stats.extract_stats(
            Neurite(root_section),
            {
                "neurite": {
                    "segment_lengths": {"modes": ["mean", "std"]},
                    "segment_meander_angles": {"modes": ["mean", "std"]},
                },
            },
        )["axon"]
        logger.debug("New trunk statistics: %s", trunk_stats)

    export(morph, initial_morph, output_path, figure_path, logger)



"""Module with tools to compute the projection intensity."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

import json
import logging
from pathlib import Path

import numpy as np
import pandas as pd
import plotly.express as px
from attrs import converters
from attrs import define
from attrs import field
from attrs import validators
from bluepyparallel import evaluate
from bluepyparallel import init_parallel_factory
from dask.distributed import LocalCluster
from morph_tool.utils import is_morphology
from neurom import COLS
from voxcell import VoxelData

from axon_synthesis.constants import FROM_COORDS_COLS
from axon_synthesis.constants import TO_COORDS_COLS
from axon_synthesis.synthesis import ParallelConfig
from axon_synthesis.typing import FileType
from axon_synthesis.utils import MorphNameAdapter
from axon_synthesis.utils import compute_bbox
from axon_synthesis.utils import disable_distributed_loggers
from axon_synthesis.utils import get_axons
from axon_synthesis.utils import load_morphology
from axon_synthesis.utils import neurite_to_pts
from axon_synthesis.validation.utils import segment_intersection_lengths

LOGGER = logging.getLogger(__name__)


def to_tuple(x: object) -> tuple:
    """Convert input data into a tuple."""
    try:
        new_x = json.loads(x)
    except Exception:
        new_x = x
    return [tuple(i) for i in new_x]


@define
class ProjectionIntensityConfig:
    """Class to store the projection intensity configuration.

    Attributes:
        morphology_dir: Path to the directory containing the input morphologies.
        grid_corner: The first corner of the grid to use.
        grid_voxel_dimensions: The voxel sizes of the grid.
        output_dir: The directory to which the results will be exported.
        figure_dir: The directory to which the figures will be exported.
    """

    morphology_dir: FileType = field(converter=Path)
    grid_voxel_dimensions: list[tuple[float, float, float]] = field(
        converter=to_tuple,
        validator=validators.deep_iterable(
            member_validator=validators.deep_iterable(
                validators.instance_of((int, float)),
                iterable_validator=validators.instance_of(tuple),
            ),
        ),
    )

    output_dir: FileType = field(converter=converters.optional(Path))
    figure_dir: FileType | None = field(default=None, converter=converters.optional(Path))


def get_morphologies(morphology_dir) -> pd.DataFrame:
    """Create a DataFrame from a directory containing morphologies."""
    morphology_dir = Path(morphology_dir)
    morph_files = [i for i in morphology_dir.iterdir() if is_morphology(i)]

    morph_names = [i.stem for i in morph_files]
    morph_files = [str(i) for i in morph_files]
    if not morph_files:
        msg = f"No morphology file found in '{morphology_dir}'"
        raise RuntimeError(msg)

    cells_df = pd.DataFrame({"morphology": morph_names, "morph_file": morph_files})
    return cells_df.sort_values("morphology", ignore_index=True)


def proj_intensities_one_morph(morph_data, config):
    """Compute the projection intensities from of a given morphology file."""
    morph_name = morph_data.get("morphology", "NO MORPH NAME FOUND")
    morph_custom_logger = MorphNameAdapter(LOGGER, extra={"morph_name": morph_name})
    file_paths = []
    grid_voxel_dimensions = []
    axon_ids = []
    try:
        # Load the morphology
        morph = load_morphology(morph_data["morph_file"])

        for num, axon in enumerate(get_axons(morph)):
            morph_custom_logger.debug("Processing axon %s", num)
            bbox = compute_bbox(axon.points[:, COLS.XYZ])
            center = axon.root_node.points[0, COLS.XYZ]

            # Create the DF of segments
            _nodes, edges = neurite_to_pts(axon, keep_section_segments=True, edges_with_coords=True)
            edges["length"] = np.linalg.norm(
                edges[FROM_COORDS_COLS].to_numpy() - edges[TO_COORDS_COLS].to_numpy(),
                axis=1,
            )

            for voxel_dimensions in config.grid_voxel_dimensions:
                # Create the grid
                heat_map = segment_intersection_lengths(edges, bbox, voxel_dimensions, center)

                # Export the result
                filename = str(
                    config.output_dir
                    / f"{morph_name}_{num}-{'_'.join([str(i) for i in voxel_dimensions])}.nrrd"
                )
                heat_map.save_nrrd(filename)
                file_paths.append(filename)
                grid_voxel_dimensions.append(voxel_dimensions)
                axon_ids.append(num)

                # Export the figure
                # if config.figure_dir is not None:
                #     plot_heat_map(
                #         heat_map,
                #         (config.figure_dir / filename.name).with_suffix(".html"),
                #     )

    except Exception:
        morph_custom_logger.exception(
            "Skip the morphology because of the following error:",
        )
        raise
    return {
        "file_paths": file_paths,
        "voxel_dimensions": grid_voxel_dimensions,
        "axon_ids": axon_ids,
    }


def compute_projection_intensities(
    config: ProjectionIntensityConfig,
    *,
    parallel_config: ParallelConfig | None = None,
):
    """Compute projection intensities of the given morphologies."""
    config.output_dir.mkdir(parents=True, exist_ok=True)
    if config.figure_dir is not None:
        config.figure_dir.mkdir(parents=True, exist_ok=True)

    # Iterate over morphologies
    morphologies = get_morphologies(config.morphology_dir)

    # Initialize parallel computation
    if parallel_config is None:
        parallel_config = ParallelConfig(0)
    with disable_distributed_loggers():
        if parallel_config.nb_processes > 1:
            LOGGER.info("Start parallel computation using %s workers", parallel_config.nb_processes)
            cluster = LocalCluster(n_workers=parallel_config.nb_processes, timeout="60s")
            parallel_factory = init_parallel_factory("dask_dataframe", address=cluster)
        else:
            LOGGER.info("Start computation")
            parallel_factory = init_parallel_factory(None)

        # Compute region indices of each segment
        results = evaluate(
            morphologies,
            proj_intensities_one_morph,
            [
                ["file_paths", None],
                ["voxel_dimensions", None],
                ["axon_ids", None],
            ],
            parallel_factory=parallel_factory,
            progress_bar=parallel_config.progress_bar,
            func_args=[config],
        )

        # Close the Dask cluster if opened
        if parallel_config.nb_processes > 1:
            parallel_factory.shutdown()
            cluster.close()

        results = (
            results.explode(["file_paths", "voxel_dimensions", "axon_ids"])
            .rename(columns={"file_paths": "file_path", "axon_ids": "axon_id"})
            .reset_index(drop=True)
        )
        results.to_hdf(config.output_dir / "files.h5", key="projection_intensity_files")

    return results


def resize_volume(volume, target_bbox):
    """Zero padding to align the volume to the target grid."""
    min_pads = np.round(np.abs(volume.bbox[0] - target_bbox[0]) / volume.voxel_dimensions)
    max_pads = np.round(np.abs(volume.bbox[1] - target_bbox[1]) / volume.voxel_dimensions)
    volume.raw = np.pad(volume.raw, np.vstack([min_pads, max_pads]).T.astype(int), "constant")
    volume.offset -= min_pads * volume.voxel_dimensions


def projection_intensities_diff(data):
    """Compare two projection intensities."""
    ref_file = data.loc["file_path_ref"]
    comp_file = data.loc["file_path_comp"]
    voxel_dimensions = data.loc["voxel_dimensions"]

    ref_data = VoxelData.load_nrrd(ref_file)
    comp_data = VoxelData.load_nrrd(comp_file)

    if (
        not np.isclose(ref_data.voxel_dimensions, voxel_dimensions).all()
        or not np.isclose(comp_data.voxel_dimensions, voxel_dimensions).all()
    ):
        msg = (
            f"Inconsistent voxel dimensions: {list(voxel_dimensions)} in data, "
            f"{list(ref_data.voxel_dimensions)} in reference file and "
            f"{list(comp_data.voxel_dimensions)} in compared file"
        )
        raise ValueError(msg)

    if not np.isclose(ref_data.offset, comp_data.offset).all() or ref_data.shape != comp_data.shape:
        LOGGER.debug("Resize data to overlap properly for %s", data.loc["morphology"])
        target_bbox = compute_bbox(np.vstack([ref_data.bbox, comp_data.bbox]))

        for i in [ref_data, comp_data]:
            resize_volume(i, target_bbox)

    diff = ref_data.with_data(ref_data.raw - comp_data.raw)

    if "file_path_diff" in data:
        diff.save_nrrd(data.loc["file_path_diff"])


def compute_projection_intensities_differences(
    ref_config: ProjectionIntensityConfig,
    compared_config: ProjectionIntensityConfig,
    output_dir: FileType,
    *,
    parallel_config: ParallelConfig | None = None,
    overwrite: bool = False,
):
    """Compute and compare the projection intensities from 2 sets of morphologies."""
    output_dir = Path(output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)

    # Compute of load the projection intensities
    if overwrite or not ref_config.output_dir.exists():
        ref = compute_projection_intensities(ref_config, parallel_config=parallel_config)
    else:
        ref = pd.read_hdf(ref_config.output_dir / "files.h5", key="projection_intensity_files")
    if overwrite or not compared_config.output_dir.exists():
        comp = compute_projection_intensities(compared_config, parallel_config=parallel_config)
    else:
        comp = pd.read_hdf(
            compared_config.output_dir / "files.h5", key="projection_intensity_files"
        )

    # Merge the data
    data = ref.merge(
        comp,
        on=["morphology", "axon_id", "voxel_dimensions"],
        suffixes=("_ref", "_comp"),
        how="outer",
        indicator=True,
    )

    # Check for inconsistent or missing data
    matching_data = data.loc[data["_merge"] == "both"].drop(columns=["_merge"])
    missing = data.loc[data["_merge"] != "both"].copy(deep=False)
    if len(missing) > 0:
        missing["found"] = (
            missing["_merge"]
            .str.replace("left_only", "in reference only")
            .str.replace("right_only", "in compared only")
        )
        LOGGER.warning(
            "Could not find corresponding entries for the following: %s",
            missing.loc[:, ["morphology", "axon_id", "voxel_dimensions", "found"]].to_dict(
                "records"
            ),
        )

    # Build the output paths
    matching_data["file_path_diff"] = matching_data.apply(
        lambda x: str(
            output_dir
            / (
                f"{x['morphology']}_{x['axon_id']}-"
                f"{'_'.join([str(i) for i in x['voxel_dimensions']])}_diff.nrrd"
            )
        ),
        axis=1,
    )

    # Export the metadata
    matching_data.to_hdf(output_dir / "metadata.h5", key="projection_intensity_differences")

    # Compute and export the diff
    matching_data.apply(projection_intensities_diff, axis=1)

    return matching_data


def all_diff_stats(data):
    """Compute basic stats of the absolute difference."""
    diff = VoxelData.load_nrrd(data.loc["file_path_diff"])
    ref = VoxelData.load_nrrd(data.loc["file_path_ref"])
    comp = VoxelData.load_nrrd(data.loc["file_path_comp"])

    # mask = np.where(ref.raw != np.nan)

    diff_masked = diff.raw  # [mask]
    ref_masked = ref.raw  # [mask]
    comp_masked = comp.raw  # [mask]

    total = np.abs(ref_masked).sum() + np.abs(comp_masked).sum()
    total_square = np.square(ref_masked).sum() + np.square(comp_masked).sum()

    return {
        "L1": np.abs(diff_masked).sum() / total,
        "L2": np.sqrt(np.square(diff_masked).sum() / total_square),
        "sum": np.abs(diff_masked).sum(),
        "mean": np.abs(diff_masked).mean(),
        "min": np.abs(diff_masked).min(),
        "max": np.abs(diff_masked).max(),
        "std": np.abs(diff_masked).std(),
    }


def diff_stats(diff_dir: FileType):
    """Compute simple statistics of projection intensity differences."""
    data = pd.read_hdf(Path(diff_dir) / "metadata.h5", key="projection_intensity_differences")
    stats = data.apply(all_diff_stats, axis=1).apply(pd.Series)
    return data.join(stats)


def plot_diff_stats(
    data,
    *,
    output_path: FileType | None = None,
    stat_type="L1",
    log_x=True,
    log_y=False,
    show=False,
):
    """Plot the projection intensity differences for the given data."""
    data = data.copy(deep=False)
    data["voxel_size"] = data["voxel_dimensions"].apply(lambda row: row[0])
    data["axon"] = data.apply(lambda row: row["morphology"] + "_" + str(row["axon_id"]), axis=1)
    fig = px.line(data, x="voxel_size", y=stat_type, color="axon", log_x=log_x, log_y=log_y)
    if output_path is not None:
        Path(output_path).parent.mkdir(parents=True, exist_ok=True)
        fig.write_html(output_path, auto_open=show)
    return fig



"""Module to define custom types used in axon-synthesis."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

# pylint: disable=unused-import
import os
import sys
from collections.abc import Sequence

import morphio
import neurom
from numpy import number as any_num
from numpy.random import BitGenerator
from numpy.random import Generator
from numpy.random import SeedSequence
from numpy.typing import ArrayLike  # noqa: F401
from numpy.typing import NDArray

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self  # noqa: F401

FileType = str | os.PathLike
LayerNamesType = list[int | str]
RegionIdsType = int | str | list[int | str]
LoadableMorphology = FileType | neurom.core.Morphology | morphio.Morphology | morphio.mut.Morphology
SeedType = None | int | Sequence[int] | SeedSequence | BitGenerator | Generator
CoordsType = list[any_num] | NDArray[any_num]



"""Some plot utils for create graph."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

from pathlib import Path

import numpy as np
import plotly.graph_objs as go

from axon_synthesis.constants import FROM_COORDS_COLS
from axon_synthesis.constants import TO_COORDS_COLS
from axon_synthesis.utils import build_layout_properties


def plot_triangulation(
    edges, source_point, target_points, figure_path, logger=None, attractors=None
):
    """Plot the given triangulation for debugging purpose."""
    segments = edges.copy(deep=False)
    segments["cutter"] = None

    fig = go.Figure()

    colors = segments["weight"] / segments["length"]
    cmin = colors.quantile(0.1)
    cmax = colors.quantile(0.9)

    edges_trace = go.Scatter3d(
        x=segments[[FROM_COORDS_COLS.X, TO_COORDS_COLS.X, "cutter"]].to_numpy().flatten().tolist(),
        y=segments[[FROM_COORDS_COLS.Y, TO_COORDS_COLS.Y, "cutter"]].to_numpy().flatten().tolist(),
        z=segments[[FROM_COORDS_COLS.Z, TO_COORDS_COLS.Z, "cutter"]].to_numpy().flatten().tolist(),
        line={
            "width": 1.5,
            "color": colors,
            "colorscale": "rainbow",
            "cmin": cmin,
            "cmax": cmax,
            "colorbar": {
                "orientation": "h",
            },
        },
        mode="lines",
        name="Steiner graph",
    )

    source_point_trace = go.Scatter3d(
        x=[source_point[0]],
        y=[source_point[1]],
        z=[source_point[2]],
        marker={"color": "purple", "size": 4},
        mode="markers",
        name="Source point",
    )

    target_points_trace = go.Scatter3d(
        x=target_points[:, 0],
        y=target_points[:, 1],
        z=target_points[:, 2],
        marker={"color": "chocolate", "size": 2},
        mode="markers",
        name="Target points",
    )

    fig.add_trace(edges_trace)
    fig.add_trace(source_point_trace)
    fig.add_trace(target_points_trace)

    pts = np.vstack([[source_point], target_points])

    if attractors is not None:
        attractors_trace = go.Scatter3d(
            x=attractors[:, 0],
            y=attractors[:, 1],
            z=attractors[:, 2],
            marker={"color": "green", "size": 3},
            mode="markers",
            name="Attractors",
        )
        fig.add_trace(attractors_trace)
        pts = np.vstack([pts, attractors])

    layout_props = build_layout_properties(pts, 0.1)

    fig.update_scenes(layout_props)
    fig.update_layout(title=Path(figure_path).stem, coloraxis_showscale=True)

    # Export figure
    Path(figure_path).parent.mkdir(parents=True, exist_ok=True)
    fig.write_html(figure_path)

    if logger is not None:
        logger.debug("Exported triangulation figure to %s", figure_path)



"""Test the examples."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

import shutil
from pathlib import Path

import pytest
from dir_content_diff import assert_equal_trees
from dir_content_diff import get_comparators
from dir_content_diff.comparators.morphio import MorphologyComparator
from dir_content_diff.comparators.pandas import HdfComparator
from dir_content_diff.comparators.voxcell import CellCollectionComparator

import axon_synthesis.cli


def _ignore_files(_src, names) -> list[str]:
    """Filter some files."""
    return [
        i for i in names if Path(i).name not in ["172992.asc", "172993.asc"] and "nested" not in i
    ]


# @pytest.mark.parametrize("nb_workers", [0, 2])
# def test_synthesis_example(testing_dir, data_dir, example_dir, cli_runner, nb_workers):
#     """Test the mimic workflow from the general example with 2 morphologies."""
#     shutil.copyfile(example_dir / "config.cfg", testing_dir / "config.cfg")
#     shutil.copyfile(example_dir / "WMR.yaml", testing_dir / "WMR.yaml")
#     shutil.copyfile(example_dir / "input_cells_lite.mvd3", testing_dir / "input_cells.mvd3")
#     shutil.copyfile(
#         example_dir / "axon_grafting_points.h5",
#         testing_dir / "axon_grafting_points.h5",
#     )

#     morph_dir = Path("morphologies") / "repair_release" / "asc"
#     shutil.copytree(example_dir / morph_dir, testing_dir / morph_dir, ignore=_ignore_files)

#     inputs_result_dir = testing_dir / "inputs"
#     result = cli_runner.invoke(
#         axon_synthesis.cli.main,
#         [
#             "-c",
#             "config.cfg",
#             "create-inputs",
#             "--output-dir",
#             str(inputs_result_dir),
#             "--nb-workers",
#             str(nb_workers),
#         ],
#     )
#     assert result.exit_code == 0, result.output

#     synthesis_result_dir = testing_dir / "out"
#     result = cli_runner.invoke(
#         axon_synthesis.cli.main,
#         [
#             "-c",
#             "config.cfg",
#             "synthesize",
#             "--input-dir",
#             inputs_result_dir,
#             "--morphology-data-file",
#             testing_dir / "input_cells.mvd3",
#             "--axon-grafting-points-file",
#             testing_dir / "axon_grafting_points.h5",
#             "--output-dir",
#             synthesis_result_dir,
#             "--nb-workers",
#             str(nb_workers),
#             "--target-max-tries",
#             "100",
#         ],
#     )
#     assert result.exit_code == 0, result.output

#     # Check the results
#     comparators = get_comparators()
#     comparators[".h5"] = MorphologyComparator(default_diff_kwargs={"rtol": 1e-3, "atol": 1e-3})
#     comparators[".asc"] = MorphologyComparator(default_diff_kwargs={"rtol": 1e-3, "atol": 1e-3})
#     out_dir_pattern = (str(synthesis_result_dir) + "/?", "")
#     assert_equal_trees()


@pytest.mark.parametrize("nb_workers", [0, 2])
def test_mimic_example(testing_dir, data_dir, example_dir, cli_runner, nb_workers):
    """Test the mimic workflow from the general example with 2 morphologies."""
    shutil.copyfile(example_dir / "config.cfg", testing_dir / "config.cfg")

    morph_dir = Path("morphologies") / "repair_release" / "asc"
    shutil.copytree(example_dir / morph_dir, testing_dir / morph_dir, ignore=_ignore_files)

    result_dir = testing_dir / "out"
    result = cli_runner.invoke(
        axon_synthesis.cli.main,
        [
            "-c",
            "config.cfg",
            "validation",
            "mimic",
            "--output-dir",
            str(result_dir),
            "--nb-workers",
            str(nb_workers),
        ],
    )
    assert result.exit_code == 0, result.output

    # Check the results
    comparators = get_comparators()
    comparators[".h5"] = MorphologyComparator(default_diff_kwargs={"rtol": 1e-3, "atol": 1e-3})
    comparators[".asc"] = MorphologyComparator(default_diff_kwargs={"rtol": 1e-3, "atol": 1e-3})
    out_dir_pattern = (str(result_dir) + "/?", "")
    assert_equal_trees(
        data_dir / "mimic_example",
        testing_dir / "out",
        comparators=comparators,
        specific_args={
            "inputs/circuit.h5": {
                "comparator": CellCollectionComparator(),
                "format_data_kwargs": {
                    "replace_pattern": {
                        out_dir_pattern: [
                            "morph_file",
                        ],
                    },
                },
            },
            "GraphCreationData data": {
                "patterns": [r"synthesis_basic_\S*/GraphCreationData/\S*\.h5$"],
                "comparator": HdfComparator(),
                "load_kwargs": {"key": "nodes"},
            },
            "SteinerTreeSolutions data": {
                "patterns": [r"synthesis_basic_\S*/SteinerTreeSolutions/\S*\.h5$"],
                "comparator": HdfComparator(),
                "load_kwargs": {"key": "solution_nodes"},
            },
            "inputs/metadata.json": {
                "format_data_kwargs": {
                    "replace_pattern": {
                        out_dir_pattern: [
                            "clustering",
                            "path",
                            "morphology_path",
                            "WMR",
                        ],
                    },
                }
            },
            "inputs/Clustering/clustered_morphologies_paths.csv": {
                "format_data_kwargs": {
                    "replace_pattern": {
                        out_dir_pattern: [
                            "morph_file",
                            "morph_path",
                        ],
                    },
                }
            },
            "inputs/Clustering/clustered_terminals.csv": {
                "format_data_kwargs": {
                    "replace_pattern": {
                        out_dir_pattern: [
                            "morph_file",
                        ],
                    },
                }
            },
            "inputs/Clustering/trunk_properties.json": {
                "tolerance": 1e-4,
                "format_data_kwargs": {
                    "replace_pattern": {
                        out_dir_pattern: [
                            "[*].morph_file",
                        ],
                    },
                },
            },
            "inputs/Clustering/tuft_properties.json": {
                "tolerance": 1e-4,
                "format_data_kwargs": {
                    "replace_pattern": {
                        out_dir_pattern: [
                            "[*].morph_file",
                        ],
                    },
                },
            },
            "inputs/projection_probabilities.csv": {
                "format_data_kwargs": {
                    "replace_pattern": {
                        out_dir_pattern: [
                            "morph_file",
                        ],
                    },
                }
            },
            "Target points": {
                "patterns": [r"synthesis_basic_\S*/target_points.h5"],
                "comparator": HdfComparator(),
                "format_data_kwargs": {
                    "replace_pattern": {
                        out_dir_pattern: [
                            "morph_file",
                        ],
                    },
                },
                "atol": 1e-6,
            },
        },
    )



"""This script creates the figure associated to the graph creation."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

import logging
from pathlib import Path

import numpy as np
import pandas as pd
import plotly.graph_objs as go
from attrs import evolve
from plotly.subplots import make_subplots
from scipy.spatial import KDTree

from axon_synthesis.constants import NodeProvider
from axon_synthesis.synthesis.main_trunk import create_graph
from axon_synthesis.synthesis.main_trunk import steiner_tree

logging.basicConfig(level=logging.DEBUG)

pd.options.display.max_rows = 100

create_graph.utils.FORCE_2D = True


def num_to_rgb(val, min_val=0, max_val=1):
    """Convert a float value to a RGB value for a rainbow color theme."""
    i = (val - min_val) * 255 / (max_val - min_val)
    coeff = np.pi / 255
    r = round(np.sin(coeff * (i - 127)) * 127 + 128)
    g = round(np.sin(coeff * (i - 63) * 2) * 127 + 128)
    b = round(np.sin(coeff * (i + 127)) * 127 + 128)

    return f"rgb({r},{g},{b})"


def degree2rad(degrees):
    """Convert degrees into radians."""
    return degrees * np.pi / 180


def disk_part(center=None, radius=1, start_angle=0, end_angle=90, n=50):
    """Create a path string for a circle arc."""
    if center is None:
        center = [0, 0]
    sta = degree2rad(start_angle)
    ea = degree2rad(end_angle)
    t = np.linspace(sta, ea, n)
    x = center[0] + radius * np.cos(t)
    y = center[1] + radius * np.sin(t)
    path = f"M {x[0]},{y[0]}"
    for xc, yc in zip(x[1:], y[1:]):
        path += f" L{xc},{yc}"
    return path


def plot(nodes, edges, figure_path, solution_edges=None, preferred_regions_pts=None):
    """Plot the given nodes and edges."""
    edges = edges.copy(deep=False)
    edges["cutter"] = None
    edges["density"] = edges["weight"] / edges["length"]

    fig = go.Figure()
    annotations = []

    common_objects = [
        go.Scatter(
            x=nodes.loc[nodes["NodeProvider"] == NodeProvider.source.name, "x"],
            y=nodes.loc[nodes["NodeProvider"] == NodeProvider.source.name, "y"],
            marker={"color": "black", "size": 25},
            mode="markers",
            name="Source point",
            legendgroup=1,
        ),
        go.Scatter(
            x=nodes.loc[nodes["NodeProvider"] == NodeProvider.target.name, "x"],
            y=nodes.loc[nodes["NodeProvider"] == NodeProvider.target.name, "y"],
            marker={"color": "rgb(255,127,0)", "size": 25},
            mode="markers",
            name="Target points",
            legendgroup=2,
        ),
        go.Scatter(
            x=nodes.loc[nodes["NodeProvider"] == NodeProvider.intermediate.name, "x"],
            y=nodes.loc[nodes["NodeProvider"] == NodeProvider.intermediate.name, "y"],
            marker={"color": "blue", "size": 20},
            mode="markers",
            name="Intermediate points",
            legendgroup=3,
        ),
        go.Scatter(
            x=nodes.loc[nodes["NodeProvider"] == NodeProvider.random.name, "x"],
            y=nodes.loc[nodes["NodeProvider"] == NodeProvider.random.name, "y"],
            marker={"color": "green", "size": 20},
            mode="markers",
            name="Random points",
            legendgroup=4,
        ),
        go.Scatter(
            x=nodes.loc[nodes["NodeProvider"] == NodeProvider.bbox.name, "x"],
            y=nodes.loc[nodes["NodeProvider"] == NodeProvider.bbox.name, "y"],
            marker={"color": "rgb(255,0,255)", "size": 20},
            mode="markers",
            name="Bounding box points",
            legendgroup=5,
        ),
        go.Scatter(
            x=nodes.loc[nodes["NodeProvider"] == NodeProvider.Voronoi.name, "x"],
            y=nodes.loc[nodes["NodeProvider"] == NodeProvider.Voronoi.name, "y"],
            marker={"color": "red", "size": 15},
            mode="markers",
            name="Voronoï points",
            legendgroup=6,
        ),
    ]
    not_preferred_objects = []
    preferred_objects = []

    if solution_edges is not None:
        solution_edges["cutter"] = None
        common_objects.append(
            go.Scatter(
                x=solution_edges[["x_from", "x_to", "cutter"]].to_numpy().flatten().tolist(),
                y=solution_edges[["y_from", "y_to", "cutter"]].to_numpy().flatten().tolist(),
                line={
                    "width": 6,
                    "color": "red",
                    "dash": "dash",
                },
                mode="lines",
                name="Steiner solution",
                legendgroup=7,
            )
        )

    if preferred_regions_pts is None:
        not_preferred_objects.append(
            go.Scatter(
                x=edges[["x_from", "x_to", "cutter"]].to_numpy().flatten().tolist(),
                y=edges[["y_from", "y_to", "cutter"]].to_numpy().flatten().tolist(),
                line={"width": 1, "color": "black"},
                mode="lines",
                name="Steiner graph edges",
                hoverinfo="text",
                legendgroup=8,
            )
        )
    else:
        edges["normalized_density"] = (edges["density"] - edges["density"].min()) / (
            edges["density"].max() - edges["density"].min()
        )
        edges["rgb_color"] = edges["normalized_density"].apply(num_to_rgb)

        for _, row in edges.iterrows():
            x = [row["x_from"], row["x_to"]]
            y = [row["y_from"], row["y_to"]]
            color = row["rgb_color"]
            preferred_objects.append(
                go.Scatter(
                    x=x,
                    y=y,
                    line={
                        "width": 2,
                        "color": color,
                    },
                    mode="lines",
                    name="Steiner graph",
                    showlegend=False,
                )
            )

        # Add empty plot just to add the color bar and the legend entry
        preferred_objects.append(
            go.Scatter(
                x=[None],
                y=[None],
                line={
                    "width": 2,
                    "color": "black",
                },
                mode="lines",
                name="Steiner graph edges",
                legendgroup=8,
            ),
        )
        annotations.append(
            {
                "text": "Linear density of edge weights",
                "font": {"size": 16, "family": "computer modern", "color": "black"},
                "showarrow": False,
                "xref": "paper",
                "yref": "paper",
                "x": 0.5,
                "y": 1.23,
            },
        )
        preferred_objects.append(
            go.Scatter(
                x=edges["x_from"],
                y=edges["y_from"],
                mode="markers",
                marker={
                    "opacity": 0,
                    "color": edges["normalized_density"],
                    "colorbar": {
                        "orientation": "h",
                        "thickness": 15,
                        "x": 0.5,
                        "y": 1,
                        "len": 0.75,
                        "ticklen": 10,
                        "tickfont": {"size": 16, "family": "computer modern", "color": "black"},
                    },
                    "colorscale": [
                        (0, "rgb(0, 0, 255)"),
                        (0.5, "rgb(127, 255, 127)"),
                        (1, "rgb(255, 0, 0)"),
                    ],
                },
                name="Steiner graph",
                showlegend=False,
                legendgroup=8,
            )
        )

        # Add the preferred region points
        for i in preferred_regions_pts:
            name = f": {i[3]}" if len(i) >= 4 and i[3] is not None else ""
            print("Add preferred region point", i)
            preferred_objects.append(
                go.Scatter(
                    x=[i[0]],
                    y=[i[1]],
                    marker={
                        "color": "maroon",
                        "size": 20,
                    },
                    mode="markers",
                    name=f"Attractor point{name}",
                    legendgroup=11,
                ),
            )
            for radius in range(0, 240, 50):
                # Add radial distances (this part of the code in not generic)
                fig.add_shape(
                    type="path",
                    xref="x",
                    yref="y",
                    path=disk_part([i[0], i[1]], radius=radius, start_angle=-90, end_angle=0),
                    x0=i[0] - radius,
                    y0=i[1] - radius,
                    x1=i[0] + radius,
                    y1=i[1] + radius,
                    line={"color": "grey", "width": 3, "dash": "dash"},
                )
            fig.add_shape(
                type="path",
                xref="x",
                yref="y",
                path=disk_part([i[0], i[1]], radius=250, start_angle=-53, end_angle=-37),
                x0=i[0] - radius,
                y0=i[1] - radius,
                x1=i[0] + radius,
                y1=i[1] + radius,
                line={"color": "grey", "width": 3, "dash": "dash"},
            )

    common_objects.append(
        go.Scatter(
            x=0.5 * (edges["x_from"] + edges["x_to"]),
            y=0.5 * (edges["y_from"] + edges["y_to"]),
            mode="markers",
            showlegend=False,
            hovertemplate="Weight of edge %{hovertext}<extra></extra>",
            hovertext=[
                f"{idx} = {weight}" for idx, weight in edges["weight"].round(2).to_dict().items()
            ],
            marker=go.scatter.Marker(opacity=0),
            legendgroup=9,
        )
    )

    fig.add_traces(common_objects)
    fig.add_traces(not_preferred_objects)
    fig.add_traces(preferred_objects)

    fig.update_layout(
        {
            "width": 600,
            "height": 500,
            "margin": {"l": 0, "r": 0, "t": 0, "b": 0, "autoexpand": True, "pad": 0},
            "paper_bgcolor": "rgba(255, 255, 255, 255)",
            "plot_bgcolor": "rgba(255, 255, 255, 255)",
            "xaxis": {
                "scaleanchor": "x",
                "scaleratio": 1,
                "showgrid": False,
                "visible": False,
            },
            "yaxis": {
                "scaleanchor": "x",
                "scaleratio": 1,
                "showgrid": False,
                "visible": False,
            },
            "legend": {
                "xanchor": "left",
                "x": 0.95,
                "yanchor": "middle",
                "y": 0.5,
                "tracegroupgap": 20,
                "font": {"size": 16, "family": "cmr10", "color": "black"},
            },
        },
        annotations=annotations,
    )

    Path(figure_path).parent.mkdir(parents=True, exist_ok=True)
    fig.write_html(figure_path + ".html")
    fig.write_image(figure_path + ".png", scale=2)
    fig.write_image(figure_path + ".svg", scale=2)

    if preferred_regions_pts is None:
        fig.update_layout({"showlegend": False})
        fig.write_html(figure_path + "_no_legend.html")
        fig.write_image(figure_path + "_no_legend.png", scale=2)
        fig.write_image(figure_path + "_no_legend.svg", scale=2)

    return fig


config = create_graph.CreateGraphConfig(
    intermediate_number=1,
    min_intermediate_distance=1,
    min_random_point_distance=50,
    random_point_bbox_buffer=0,
    voronoi_steps=1,
    use_orientation_penalty=False,
    use_depth_penalty=False,
    use_terminal_penalty=True,
)
source_coords = np.array([0, 0, 0])
target_points = pd.DataFrame(
    {
        "axon_id": [1, 1],
        "terminal_id": [0, 1],
        "target_x": [200, 200],
        "target_y": [-100, 100],
        "target_z": [0, 0],
    }
)
target_points = target_points.astype(
    dtype={"target_x": float, "target_y": float, "target_z": float}
)

# Case without preferred region
nodes, edges = create_graph.one_graph(source_coords, target_points, config, rng=1)

# ################################################################################ #
# Let's trick the approximation a bit to ensure the solution is optimal for having #
# a more understandable figure. This is required because the approximation can't   #
# find the optimal path in this specific case.                                     #
edges.loc[[5, 38, 54, 58, 59], "weight"] = 1
# ################################################################################ #

_, solution_edges = steiner_tree.compute_solution(
    nodes,
    edges,
)


graph_creation_fig = plot(nodes, edges, "graph_creation")
graph_creation_solution_fig = plot(nodes, edges, "graph_creation_solution", solution_edges)

# ruff: noqa: T201
print("Nodes:")
print(nodes)
print("Edges:")
print(edges)
print("Solution total length:", edges.loc[edges["is_solution"], "length"].sum())


# Case with preferred region
points_preferred_regions = [[0, 100, 0]]
config.preferred_region_tree = KDTree(points_preferred_regions)
preferred_regions_config = evolve(
    config,
    preferring_sigma=100,
    preferring_amplitude=10,
    use_orientation_penalty=False,
    use_depth_penalty=False,
    use_terminal_penalty=False,
)

nodes_preferred_regions, edges_preferred_regions = create_graph.one_graph(
    source_coords, target_points, preferred_regions_config, rng=1
)

_, solution_edges_preferred_regions = steiner_tree.compute_solution(
    nodes_preferred_regions,
    edges_preferred_regions,
)

preferred_region_fig = plot(
    nodes_preferred_regions,
    edges_preferred_regions,
    "graph_creation_preferred_regions",
    preferred_regions_pts=points_preferred_regions,
)
preferred_region_solution_fig = plot(
    nodes_preferred_regions,
    edges_preferred_regions,
    "graph_creation_solution_preferred_regions",
    solution_edges_preferred_regions,
    preferred_regions_pts=points_preferred_regions,
)

# Combine the two figures into one with subplots
combined_fig = make_subplots(rows=1, cols=2, horizontal_spacing=0)
for i in graph_creation_solution_fig.data:
    i.showlegend = False
    combined_fig.add_trace(i, row=1, col=1)
for i in preferred_region_solution_fig.data:
    combined_fig.add_trace(i, row=1, col=2)
annotations = list(preferred_region_solution_fig.select_annotations())
combined_fig.update_layout(
    {
        "width": 1000,
        "height": 650,
        "margin": {"l": 0, "r": 0, "t": 0, "b": 0, "autoexpand": True, "pad": 0},
        "paper_bgcolor": "rgba(255, 255, 255, 255)",
        "plot_bgcolor": "rgba(255, 255, 255, 255)",
        "xaxis1": {
            "scaleanchor": "x",
            "scaleratio": 1,
            "showgrid": False,
            "visible": False,
        },
        "yaxis1": {
            "scaleanchor": "x",
            "scaleratio": 1,
            "showgrid": False,
            "visible": False,
        },
        "xaxis2": {
            "scaleanchor": "x",
            "scaleratio": 1,
            "showgrid": False,
            "visible": False,
        },
        "yaxis2": {
            "scaleanchor": "x",
            "scaleratio": 1,
            "showgrid": False,
            "visible": False,
        },
        "legend": {
            "entrywidth": 250,
            "xanchor": "center",
            "x": 0.5,
            "yanchor": "top",
            "y": 0,
            "tracegroupgap": 10,
            "font": {"size": 16, "family": "cmr10", "color": "black"},
            "orientation": "h",
        },
    },
    annotations=annotations,
)
figure_path = "graph_creation_solution_preferred_regions_subplots"
combined_fig.write_html(figure_path + ".html")
combined_fig.write_image(figure_path + ".png", scale=2)
combined_fig.write_image(figure_path + ".svg", scale=2)
